"""
Agent
=====

We follow registry pattern for models and actions.
"""

import copy
from uuid import uuid4
from typing import Any, List, Optional, Dict, Tuple

import jinja2

from chainfury.utils import logger
from chainfury.base import (
    func_to_vars,
    func_to_return_vars,
    extract_jinja_indices,
    get_value_by_keys,
    put_value_by_keys,
    Node,
    Model,
    Var,
)

# Models
# ------
# All the things below are for the models that are registered in the model registry, so that they can be used as inputs
# in the chain. There can be several models that can put as inputs in a single chatbot.


class ModelRegistry:
    """Model registry contains metadata for all the models that are provided in the components"""

    def __init__(self):
        self.models: Dict[str, Model] = {}
        self.counter: Dict[str, int] = {}
        self.tags_to_models: Dict[str, List[str]] = {}

    def has(self, id: str):
        """A helper function to check if a model is registered or not"""
        return id in self.models

    def register(self, model: Model):
        """Register a model in the registry

        Args:
            model (Model): Model to register
        """
        id = f"{model.id}"
        logger.debug(f"Registering model {id} at {id}")
        if id in self.models:
            raise Exception(f"Model {id} already registered")
        self.models[id] = model
        for tag in model.tags:
            self.tags_to_models[tag] = self.tags_to_models.get(tag, []) + [id]

    def get_tags(self) -> List[str]:
        """Get all the tags that are registered in the registry

        Returns:
            List[str]: List of tags
        """
        return list(self.tags_to_models.keys())

    def get_models(self, tag: str = "") -> Dict[str, Dict[str, Any]]:
        """Get all the models that are registered in the registry

        Args:
            tag (str, optional): Filter models by tag. Defaults to "".

        Returns:
            Dict[str, Dict[str, Any]]: Dictionary of models
        """
        items = {k: v.to_dict() for k, v in self.models.items()}
        if tag:
            items = {k: v for k, v in items.items() if tag in v.get("tags", [])}
        return items

    def get(self, id: str) -> Model:
        """Get a model from the registry

        Args:
            id (str): Id of the model

        Returns:
            Model: Model
        """
        self.counter[id] = self.counter.get(id, 0) + 1
        out = self.models.get(id, None)
        if out is None:
            raise ValueError(f"Model {id} not found")
        return out

    def get_count_for_model(self, id: str) -> int:
        """Get the number of times a model is used

        Args:
            id (str): Id of the model

        Returns:
            int: Number of times the model is used
        """
        return self.counter.get(id, 0)


model_registry = ModelRegistry()
"""
`model_registry` is a global variable that is used to register models. It is an instance of ModelRegistry class.
"""


# Programtic Actions Registry
# ---------------------------
# Programtic actions are nodes that are software 1.0 nodes, i.e. they are not trainable. They are used for things like
# calling an API, adding 2 numbers, etc. Since they are not trainable the only way to get those is the source code for
# the server.


class ProgramaticActionsRegistry:
    """Programatic actions registry contains metadata for all the programatic actions that are provided in the
    components"""

    def __init__(self):
        self.nodes: Dict[str, Node] = {}
        self.counter: Dict[str, int] = {}
        self.tags_to_nodes: Dict[str, List[str]] = {}

    def register(
        self,
        fn: object,
        node_id: str,
        description: str,
        returns: List[str] = [],
        outputs=None,
        tags: List[str] = [],
    ) -> Node:
        """Register a programatic action in the registry

        Args:
            fn (object): Function to register
            node_id (str): Id of the node
            description (str): Description of the node
            returns (List[str], optional): List of returns. Defaults to [].
            outputs ([type], optional): [description]. Defaults to None.
            tags (List[str], optional): List of tags. Defaults to [].

        Raises:
            Exception: If the node is already registered

        Returns:
            Node: Node
        """
        logger.debug(f"Registering p-node '{node_id}'")
        if node_id in self.nodes:
            raise Exception(f"Node '{node_id}' already registered")
        if not outputs:
            assert len(returns), "If outputs is not provided then returns must be provided"
            outputs = {x: () for x in returns}
        else:
            assert len(outputs), "If returns is not provided then outputs must be provided"
        ops = func_to_return_vars(func=fn, returns=outputs)
        node = Node(
            id=node_id,
            type=Node.types.PROGRAMATIC,
            fn=fn,
            description=description,
            fields=func_to_vars(fn),
            outputs=ops,
            tags=tags,
        )
        self.nodes[node_id] = node
        for tag in tags:
            self.tags_to_nodes[tag] = self.tags_to_nodes.get(tag, []) + [node_id]
        return self.nodes[node_id]

    def get_tags(self) -> List[str]:
        """Get all the tags that are registered in the registry

        Returns:
            List[str]: List of tags
        """
        return list(self.tags_to_nodes.keys())

    def get_nodes(self, tag: str = "") -> Dict[str, Dict[str, Any]]:
        """Get all the nodes that are registered in the registry

        Args:
            tag (str, optional): Filter nodes by tag. Defaults to "".

        Returns:
            Dict[str, Dict[str, Any]]: Dictionary of nodes
        """
        items = {k: v.to_dict() for k, v in self.nodes.items()}
        if tag:
            items = {k: v for k, v in items.items() if tag in v.get("tags", [])}
        return items

    def get(self, node_id: str) -> Optional[Node]:
        """Get a node from the registry

        Args:
            node_id (str): Id of the node

        Returns:
            Node: Node
        """
        self.counter[node_id] = self.counter.get(node_id, 0) + 1
        out = self.nodes.get(node_id, None)
        if out is None:
            raise ValueError(f"p-node '{node_id}' not found")
        return copy.deepcopy(out)

    def get_count_for_nodes(self, node_id: str) -> int:
        """Get the number of times a node is used

        Args:
            node_id (str): Id of the node

        Returns:
            int: Number of times the node is used
        """
        return self.counter.get(node_id, 0)


programatic_actions_registry = ProgramaticActionsRegistry()
"""
`programatic_actions_registry` is a global variable that is used to register programatic nodes. It is an instance of
ProgramaticActionsRegistry class.
"""


# AI Actions Registry
# -------------------
# For everything that cannot be done by we have the AI powered actions Registry. This registry
# will not include all the things that are available to the outer service, but those that are
# hardcoded in the entire thing somewhere.


class AIAction:
    """This class is a callable for all the AI actions.

    Args:
        node_id (str): The id of the node
        model (Model): The model that is used for this action
        model_params (Dict[str, Any]): The parameters for the model
        fn (object): The function that is used for this action
    """

    # do not remove these from here it is used in base.py if you do then put in a third file
    JTYPE = "jinja-template"
    """constant for Jinja template type"""

    FUNC = "python-function"
    """constant for Python function type"""

    def __init__(self, node_id: str, model: Model, model_params: Dict[str, Any], fn: object, action_name: str):
        # do some basic checks that we can do before anything else like checking if model_params
        # is a subset of the model.vars
        fields = set(x.name for x in model.vars)
        mp_set = set(model_params.keys())
        if not mp_set.issubset(fields):
            raise Exception(f"Model params {mp_set} not a subset of {fields}")

        self.templates = []

        # since this is the AI action this is responsible for validating the function
        if type(fn) == dict:
            action_source = AIAction.JTYPE
            fields = []  # fields required for self.fields
            templates = []  # list of all the templates to be render with its position in fn
            fields_with_locations = extract_jinja_indices(fn)
            for field in fields_with_locations:
                fields.extend(field[1])
                obj = get_value_by_keys(fn, field[0])
                if not obj:
                    raise ValueError(f"Field {field[0]} not found in {fn}, but was extraced. There is a bug in get_value_by_keys function")
                templates.append((obj, jinja2.Template(obj), field[0]))

            # set values
            self.templates = templates
        else:
            assert type(fn) == type(func_to_return_vars), "`fn` can either be a function or a string"
            action_source = AIAction.FUNC
            fields = func_to_vars(fn)

        self.node_id = node_id
        self.model = model
        self.model_params = model_params
        self.fn = fn
        self.action_name = action_name
        self.action_source = action_source
        self.fields = fields

    def to_dict(self, no_vars: bool = False) -> Dict[str, Any]:
        """Serialize the AIAction object to a dict."""
        return {
            "node_id": self.node_id,
            "model": self.model.to_dict(no_vars=no_vars),
            "model_params": self.model_params,
            "fn": self.fn,
            "action_name": self.action_name,
            "action_source": self.action_source,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """Deserialize the AIAction object from a dict."""
        return cls(
            node_id=data["node_id"],
            model=model_registry.get(data["model"]["id"]),
            model_params=data["model_params"],
            fn=data["fn"],
            action_name=data.get("action_name", data["node_id"]),
        )

    def __call__(self, **data: Dict[str, Any]) -> Tuple[Any, Optional[Exception]]:
        """This is a callable that takes in all the arguments that the underlying models take.

        Args:
            **data (Dict[str, Any]): The data that is passed to the model

        Returns:
            Tuple[Any, Optional[Exception]]: The output of the model and the exception if any
        """

        # check for keys even before calling any API or something
        # we need to create a sub dict that only contains the fields that are needed by the preprocessor
        # function and pass the rest of the data to the model call
        _data = {}
        for f in self.fields:
            if f.required and f.name not in data:
                raise Exception(f"Field '{f.name}' is required in {self.node_id} but not present")
            if f.name in data:
                _data[f.name] = data.pop(f.name)

        if self.action_source == AIAction.FUNC:
            try:
                fn_out = self.fn(**_data)  # type: ignore
                if self.action_source == AIAction.FUNC and not type(fn_out) == dict:
                    raise Exception(f"AI Action preprocessor for {self.node_id} did not return a dict but {type(fn_out)}")
            except Exception as e:
                return "", e
        elif self.action_source == AIAction.JTYPE:
            fn_out = copy.deepcopy(self.fn)
            for raw, t, keys in self.templates:
                value = t.render(**_data)
                put_value_by_keys(fn_out, keys, value)

        # print(">> model_params:", self.model_params)
        # print(">> preprocessor:", fn_out)
        model_final_params = {**self.model_params}
        model_final_params.update(data)
        model_final_params.update(fn_out)  # type: ignore
        out, err = self.model(model_final_params)
        if err != None:
            return "", err

        return out, err


class AIActionsRegistry:
    """This class is a registry for all the AI actions."""

    DB_REGISTER = "db"

    def __init__(self):
        self.nodes: Dict[str, Node] = {}
        self.counter: Dict[str, int] = {}
        self.tags_to_nodes: Dict[str, List[str]] = {}

    def to_action(
        self,
        action_name: str,
        model_id: str,
        model_params: Dict[str, Any],
        fn: object,
        outputs: Dict[str, Any],
        node_id: str = "",
        description: str = "",
    ) -> Node:
        """
        function to create an "Action" aka. `chainfury.Node`.

        **NOTE:** If you do not pass `node_id` then this will create a `uudi4`. This behaviour is important when dev
        wants to play with the action without commiting it anywhere.

        Args:
            model_id (str): The model id that is to be used for this action
            model_params (Dict[str, Any]): The model params that are to be used for this action
            fn (object): The function that is to be used as a preprocessor for this action
            outputs: This is a dict like `{'x': (-1, 'b', 'c')}`, if provided function returns a dictionary with key `x`
              and value automatically extracted from the model output at location `(-1, 'b', 'c')`.
            node_id (str, optional): The node id for this action. Defaults to "".
            description (str, optional): The description for this action. Defaults to "".

        Returns:
            Node: The node object that can be used to create a chain
        """
        node_id = node_id or str(uuid4())
        model = model_registry.get(model_id)
        if model is None:
            raise Exception(f"Model {model_id} not found")
        ai_action = AIAction(
            node_id=node_id,
            model=model,
            model_params=model_params,
            fn=fn,
            action_name=action_name,
        )
        if not outputs:
            output_field = func_to_return_vars(func=ai_action.__call__, returns={"model_output": ()})
        else:
            output_field = [Var(type="string", name=k, loc=loc) for k, loc in outputs.items()]
        node = Node(
            id=node_id,
            fn=ai_action,
            type=Node.types.AI,
            description=description,
            fields=ai_action.fields + model.vars,
            outputs=output_field,
        )
        return node

    def register(
        self,
        node_id: str,
        model_id: str,
        model_params: Dict[str, Any],
        fn: object,
        outputs: Dict[str, Any],
        action_name: str = "",
        description: str = "",
        tags: List[str] = [],
    ) -> Node:
        """
        This function will register this action in the local AI registry so it is accesible everywhere. Use this when
        you are hosting your own chainfury server and want to serve private functions not available in the public DB.
        If you are using this just for your local usecase and have no interest in the serving capabilities then
        use `to_action` instead.

        **NOTE:** If you do not pass `node_id` then this will create a `uudi4`. This behaviour is important when dev
        wants to play with the action without commiting it anywhere.

        Args:
            node_id (str): The node id for this action
            model_id (str): The model id that is to be used for this action
            model_params (Dict[str, Any]): The model params that are to be used for this action
            fn (object): The function that is to be used as a preprocessor for this action
            outputs: This is a dict like `{'x': (-1, 'b', 'c')}`, if provided function returns a dictionary with key `x`
                and value automatically extracted from the model output at location `(-1, 'b', 'c')`.
            description (str, optional): The description for this action. Defaults to "".
            tags (List[str], optional): The tags for this action. Defaults to [].
        """
        logger.debug(f"Registering ai-node '{node_id}'")
        if node_id != AIActionsRegistry.DB_REGISTER and node_id in self.nodes:
            raise ValueError(f"ai-node '{node_id}' already exists")
        node = self.to_action(
            action_name=action_name or node_id,
            node_id=node_id,
            model_id=model_id,
            model_params=model_params,
            fn=fn,
            outputs=outputs,
            description=description,
        )

        # this node can be registered in DB
        if node_id == AIActionsRegistry.DB_REGISTER:
            pass

        # this is just the server instance register
        else:
            self.nodes[node_id] = node
            for tag in tags:
                self.tags_to_nodes[tag] = self.tags_to_nodes.get(tag, []) + [node_id]
        return node

    def unregister(self, node_id: str):
        """Unregister an ai-node

        Args:
            node_id (str): The node id for this action

        Raises:
            ValueError: If the node is not found
        """
        logger.debug(f"Unregistering ai-node '{node_id}'")
        node = self.nodes.pop(node_id, None)
        if node is None:
            raise ValueError(f"ai-node '{node_id}' not found")
        for tag, nodes in self.tags_to_nodes.items():
            if node_id in nodes:
                nodes.remove(node_id)
                if not nodes:
                    self.tags_to_nodes.pop(tag)
                break

    def get_tags(self) -> List[str]:
        """Get all the tags that are registered

        Returns:
            List[str]: The list of tags
        """
        return list(self.tags_to_nodes.keys())

    def get_nodes(self, tag: str = "") -> Dict[str, Dict[str, Any]]:
        """Get all the nodes that are registered

        Args:
            tag (str, optional): The tag to filter the nodes. Defaults to "".

        Returns:
            Dict[str, Dict[str, Any]]: The dict of nodes
        """
        items = {k: v.to_dict() for k, v in self.nodes.items()}
        if tag:
            items = {k: v for k, v in items.items() if tag in v.get("tags", [])}
        return items

    def get(self, node_id: str) -> Optional[Node]:
        """Get the node for the given node id

        Args:
            node_id (str): The node id for this action

        Returns:
            Optional[Node]: The node object
        """
        self.counter[node_id] = self.counter.get(node_id, 0) + 1
        out = self.nodes.get(node_id, None)
        if out is None:
            raise ValueError(f"ai-node '{node_id}' not found")
        return Node.from_dict(copy.deepcopy(out.to_dict()))

    def get_count_for_nodes(self, node_id: str) -> int:
        """Get number of times a particular node is called

        Args:
            node_id (str): The node id for this action

        Returns:
            int: The number of times the node is called
        """
        return self.counter.get(node_id, 0)


ai_actions_registry = AIActionsRegistry()
"""
`ai_actions_registry` is a global instance of `AIActionsRegistry` class. This is used to register and unregister
`AIAction` instances. This is used by the server to serve the registered actions.
"""

DEFAULT_MEMORY_CONSTANTS = {
    "openai-embedding": {
        "embedding_model_key": "input_strings",
        "embedding_model_params": {
            "model": "text-embedding-ada-002",
        },
        "translation_layer": {
            "embeddings": ["data", "*", "embedding"],
        },
    }
}


class Memory:
    """Class to wrap the DB functions as a callable.

    Args:
        node_id (str): The id of the node
        fn (object): The function that is used for this action
        vector_key (str): The key for the vector in the DB
        read_mode (bool, optional): If the function is a read function, if `False` then this is a write function.
    """

    fields_model = [
        Var(name="items", type=[Var(type="string"), Var(type="array", items=[Var(type="string")])], required=True),
        Var(name="embedding_model", type="string", required=True),
        Var(name="embedding_model_params", type="object", additionalProperties=Var(type="string")),
        Var(name="embedding_model_key", type="string"),
        Var(name="translation_layer", type="object", additionalProperties=Var(type="string")),
    ]
    """These are the fields that are used to map the input items to the embedding model, do not use directly"""

    def __init__(self, node_id: str, fn: object, vector_key: str, read_mode: bool = False):
        self.node_id = node_id
        self.fn = fn
        self.vector_key = vector_key
        self.read_mode = read_mode
        self.fields_fn = func_to_vars(fn)
        self.fields = self.fields_fn + self.fields_model

    def to_dict(self) -> Dict[str, Any]:
        """Serialize the Memory object to a dict."""
        return {
            "node_id": self.node_id.split("-")[0],
            "vector_key": self.vector_key,
            "read_mode": self.read_mode,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """Deserialize the Memory object from a dict."""
        read_mode = data["read_mode"]
        if read_mode:
            fn = memory_registry.get_read(data["node_id"])
        else:
            fn = memory_registry.get_write(data["node_id"])
        return fn.fn

    def __call__(self, **data: Dict[str, Any]) -> Any:
        # the first thing we have to do is get the data for the model. This is actually a very hard problem because this
        # function needs to call some other arbitrary function where we know the inputs to this function "items" but we
        # do not know which variable to pass this to in the undelying model's function. Thus we need to take in a huge
        # amount of things as more inputs ("embedding_model_key", "embedding_model_params"). Then we don't even know
        # what the inputs to the underlying DB functionbare going to be, in which case we also need to add things like
        # the translation that needs to be done ("translation_layer"). This makes the number of inputs a lot but
        # ultimately is required to do the job for robust-ness. Which is why we provide a default for openai-embedding
        # model. For any other model user will need to pass all the information.
        model_fields: Dict[str, Any] = {}
        for f in self.fields_model:
            if f.required and f.name not in data:
                raise Exception(f"Field '{f.name}' is required in {self.node_id} but not present")
            if f.name in data:
                model_fields[f.name] = data.pop(f.name)

        model_data = {**model_fields.get("embedding_model_params", {})}
        model_id = model_fields.pop("embedding_model")
        embedding_model_default_config = DEFAULT_MEMORY_CONSTANTS.get(model_id, {})
        if embedding_model_default_config:
            model_data = {**embedding_model_default_config.get("embedding_model_params", {}), **model_data}
            model_key = embedding_model_default_config.get("embedding_model_key", "items") or model_data.get("embedding_model_key")
            model_fields["translation_layer"] = model_fields.get("translation_layer") or embedding_model_default_config.get(
                "translation_layer"
            )
        else:
            req_keys = [x.name for x in self.fields_model[2:]]
            if not all([x in model_fields for x in req_keys]):
                raise Exception(f"Model {model_id} requires {req_keys} to be passed")
            model_key = model_fields.get("embedding_model_key")
            model_data = {**model_fields.get("embedding_model_params", {}), **model_data}
        model_data[model_key] = model_fields.pop("items")  # type: ignore
        model = model_registry.get(model_id)
        embeddings, err = model(model_data=model_data)
        if err:
            logger.error(f"error: {err}")
            logger.error(f"traceback: {embeddings}")
            raise err

        # now that we have all the embeddings ready we now need to translate it to be fed into the DB function
        translated_data = {}
        for k, v in model_fields.get("translation_layer", {}).items():
            translated_data[k] = get_value_by_keys(embeddings, v)

        # create the dictionary to call the underlying function
        db_data = {}
        for f in self.fields_fn:
            if f.required and not (f.name in data or f.name in translated_data):
                raise Exception(f"Field '{f.name}' is required in {self.node_id} but not present")
            if f.name in data:
                db_data[f.name] = data.pop(f.name)
            if f.name in translated_data:
                db_data[f.name] = translated_data.pop(f.name)
        out, err = self.fn(**db_data)  # type: ignore
        return out, err


class MemoryRegistry:
    def __init__(self) -> None:
        self._memories = {}

    def register_write(
        self,
        component_name: str,
        fn: object,
        outputs: Dict[str, Any],
        vector_key: str,
        description: str = "",
        tags: List[str] = [],
    ) -> Node:
        node_id = f"{component_name}-write"
        mem_fn = Memory(node_id=node_id, fn=fn, vector_key=vector_key, read_mode=False)
        output_fields = func_to_return_vars(fn, returns=outputs)
        node = Node(
            id=node_id,
            fn=mem_fn,
            type=Node.types.MEMORY,
            fields=mem_fn.fields,
            outputs=output_fields,
            description=description,
            tags=tags,
        )
        self._memories[node_id] = node
        return node

    def register_read(
        self,
        component_name: str,
        fn: object,
        outputs: Dict[str, Any],
        vector_key: str,
        description: str = "",
        tags: List[str] = [],
    ) -> Node:
        node_id = f"{component_name}-read"
        mem_fn = Memory(node_id=node_id, fn=fn, vector_key=vector_key, read_mode=True)
        output_fields = func_to_return_vars(fn, returns=outputs)
        node = Node(
            id=node_id,
            fn=mem_fn,
            type=Node.types.MEMORY,
            fields=mem_fn.fields,
            outputs=output_fields,
            description=description,
            tags=tags,
        )
        self._memories[node_id] = node
        return node

    def get_write(self, node_id: str) -> Optional[Node]:
        out = self._memories.get(node_id + "-write", None)
        if out is None:
            raise ValueError(f"Memory '{node_id}' not found")
        return out

    def get_read(self, node_id: str) -> Optional[Node]:
        out = self._memories.get(node_id + "-read", None)
        if out is None:
            raise ValueError(f"Memory '{node_id}' not found")
        return out

    def get_nodes(self):
        return {k: v.to_dict() for k, v in self._memories.items()}


memory_registry = MemoryRegistry()
"""
`memory_registry` is a global instance of MemoryRegistry class. This is used to register and unregister Memory instances.
This is what the user should use when they want to use the memory elements in their chain.
"""



import os
import sys
import json
from fire import Fire

from chainfury import Chain
from chainfury.utils import logger
from chainfury.client import get_client
from chainfury.version import __version__
from chainfury.components import all_items
from chainfury.agent import model_registry, programatic_actions_registry, memory_registry


def help():
    print(
        """
🦋 Welcome to ChainFury Engine!

A powerful way to program for the "Software 2.0" era. Read more:

- https://blog.nimblebox.ai/new-flow-engine-from-scratch
- https://blog.nimblebox.ai/fury-actions
- https://gist.github.com/yashbonde/002c527853e04869bfaa04646f3e0974

🌟 us on https://github.com/NimbleBoxAI/ChainFury

Build with ♥️  by NimbleBox.ai

🌊 Chennai, India
"""
    )


def run(
    chain: str,
    inp: str,
    stream: bool = False,
    print_thoughts: bool = False,
    f=sys.stdout,
):
    """
    Run a chain with input and write the outputs.

    Args:
        chain (str): This can be one of json filepath (e.g. "/chain.json"), json string (e.g. '{"id": "99jcjs9j2", ...}'),
            chain id (e.g. "99jcjs9j2")
        inp (str): This can be one of json filepath (e.g. "/input.json"), json string (e.g. '{"foo": "bar", ...}')
        stream (bool, optional): Whether to stream the output. Defaults to False.
        print_thoughts (bool, optional): Whether to print thoughts. Defaults to False.
        f (file, optional): File to write the output to. Defaults to `sys.stdout`.

    Examples:
        >>> $ cf run ./sample.json {"foo": "bar"}
    """
    # validate inputs
    if isinstance(inp, str):
        if os.path.exists(inp):
            with open(inp, "r") as f:
                inp = json.load(f)
        else:
            inp = json.loads(inp)
    assert isinstance(inp, dict), "Input must be a dict"

    # create chain
    chain_obj = None
    if isinstance(chain, str):
        if os.path.exists(chain):
            with open(chain, "w") as f:
                chain = json.load(f)
        if len(chain) == 8:
            chain_obj = Chain.from_id(chain)
        else:
            chain = json.loads(chain)
    elif isinstance(chain, dict):
        chain_obj = Chain.from_dict(chain)
    assert chain_obj is not None, "Chain not found"

    # output
    if isinstance(f, str):
        f = open(f, "w")

    # run the chain
    if stream:
        cf_response_gen = chain_obj.stream(inp, print_thoughts=print_thoughts)
        for ir, done in cf_response_gen:
            if not done:
                f.write(json.dumps(ir) + "\n")
    else:
        out, buffer = chain_obj(inp, print_thoughts=print_thoughts)
        for k, v in buffer.items():
            f.write(json.dumps({k: v}) + "\n")

    # close file
    f.close()


def main():
    Fire(
        {
            "comp": {
                "all": lambda: print(all_items),
                "model": {
                    "list": list(model_registry.get_models()),
                    "all": model_registry.get_models(),
                },
                "prog": {
                    "list": list(programatic_actions_registry.get_nodes()),
                    "all": programatic_actions_registry.get_nodes(),
                },
                "memory": {
                    "list": list(memory_registry.get_nodes()),
                    "all": memory_registry.get_nodes(),
                },
            },
            "help": help,
            "run": run,
            "version": lambda: print(
                f"""ChainFury 🦋 Engine

chainfury=={__version__}
"""
            ),
        }
    )



import os
import requests
from functools import lru_cache
from typing import Dict, Any, Tuple

from chainfury.utils import logger
from chainfury.base import Chain, Node, Edge
from chainfury.agent import ai_actions_registry, programatic_actions_registry
from chainfury.types import Dag, FENode, Edge as EdgeType


class Subway:
    """
    Simple code that allows writing APIs by `.attr.ing` them. This is inspired from gRPC style functional calls which
    hides the complexity of underlying networking. This is useful when you are trying to debug live server directly.

    Note:
        User is solely responsible for checking if the certain API endpoint exists or not. This simply wraps the API
        calls and does not do any validation.

    Example:
        >>> from chainfury.client import Subway
        >>> from requests import Session
        >>> session = Session()
        >>> session.headers.update({"token": token})
        >>> stub = Subway("http://localhost:8000", session)
        >>> get_chain = stub.chatbot.u("6ln9ksln")       # http://localhost:8000/chatbot/6ln9ksln
        >>> chain = get_chain()                          # call like a function
        {
            'name': 'funny-bot-1',
            'description': None,
            'dag': {
                'nodes': [
                    {
                        'id': 'bc1bdc37-07d9-49b4-9e09-b0e58a535da5_934.2328674347034',
                        'cf_id': 'bc1bdc37-07d9-49b4-9e09-b0e58a535da5',
                        'position': {'x': -271.25233176301117, 'y': 78.20693852768798},
                        'type': 'FuryEngineNode',
                        'width': 350,
                        'height': 553,
                        'selected': True,
                        'position_absolute': None,
                        'dragging': False,
                        'data': {}
                    }
                ],
                'edges': [],
                'sample': {
                    'bc1bdc37-07d9-49b4-9e09-b0e58a535da5_934.2328674347034/model': 'gpt-3.5-turbo'
                },
                'main_in': 'bc1bdc37-07d9-49b4-9e09-b0e58a535da5_934.2328674347034/animal',
                'main_out': 'bc1bdc37-07d9-49b4-9e09-b0e58a535da5_934.2328674347034/text'
            },
            'engine': 'fury',
            'deleted_at': None,
            'created_by': 'cihua4hh',
            'id': '6ln9ksln',
            'meta': None,
            'created_at': '2023-06-27T18:05:17.395260'
        }

    Args:
        _url (str): The url to use for the client
        _session (requests.Session): The session to use for the client
    """

    def __init__(self, _url, _session):
        self._url = _url.rstrip("/")
        self._session = _session

    def __repr__(self):
        return f"<Subway ({self._url})>"

    def __getattr__(self, attr: str):
        # https://stackoverflow.com/questions/3278077/difference-between-getattr-vs-getattribute
        return Subway(f"{self._url}/{attr}", self._session)

    def u(self, attr: str) -> "Subway":
        """In cases where the api might start with a number you cannot write in python, this method can be used to
        access the attribute.

        Example:
            >>> stub.9jisjfi      # python will cry, invalid syntax: cannot start with a number
            >>> stub.u('9jisjfi') # do this instead

        Args:
            attr (str): The attribute to access

        Returns:
            Subway: The new subway object
        """
        return getattr(self, attr)

    def __call__(
        self,
        method="get",
        trailing="",
        json={},
        data=None,
        params: Dict = {},
        _verbose=False,
        **kwargs,
    ) -> Tuple[Dict[str, Any], bool]:
        """Call the API endpoint as if it is a function.

        Args:
            method (str, optional): The method to use. Defaults to "get".
            trailing (str, optional): The trailing url to use. Defaults to "".
            json (Dict[str, Any], optional): The json to use. Defaults to {}.
            data ([type], optional): The data to use. Defaults to None.
            params (Dict, optional): The params to use. Defaults to {}.
            _verbose (bool, optional): Whether to print the response or not. Defaults to False.

        Returns:
            Tuple[Dict[str, Any], bool]: The response and whether there was an error or not
        """
        fn = getattr(self._session, method)
        url = f"{self._url}{trailing}"
        if _verbose:
            logger.info(f"Calling {url}")
        items = {}
        if json:
            items["json"] = json
        if data:
            items["data"] = data
        if params:
            items["params"] = params
        r = fn(url, **items, **kwargs)
        if _verbose:
            logger.info(r.content.decode())
        try:
            r.raise_for_status()  # good when server is good
            return r.json(), False
        except:
            return r.content.decode(), True


@lru_cache(maxsize=1)
def get_client(prefix: str = "api/v1", url="", token: str = "") -> Subway:
    """This function returns a Subway object that can be used to interact with the API.

    Example:
        >>> from chainfury import get_client
        >>> client = get_client()
        >>> cf_actions = client.api.v1.fury.actions.list()
        >>> cf_actions

    Note:
        The `get_client` function is a convenience function that can be used to get a client object. It is not required
        to use the library. Under the hood, it still will call the chainfury REST endpoints.

    Args:
        prefix (str, optional): The prefix to use for the client. Defaults to "api/v1".
        url (str, optional): The url to use for the client or picks from `CF_URL` env var. Defaults to "".
        token (str, optional): The token to use for the client or picks from `CF_TOKEN` env var. Defaults to "".

    Raises:
        ValueError: If no url or token is provided.

    Returns:
        Subway: A Subway object that can be used to interact with the API.
    """
    url = url or os.environ.get("CF_URL", "")
    if not url:
        raise ValueError("No url provided, please set CF_URL environment variable or pass url as argument")
    token = token or os.environ.get("CF_TOKEN", "")
    if not token:
        raise ValueError("No token provided, please set CF_TOKEN environment variable or pass token as argument")

    session = requests.Session()
    session.headers.update({"token": token})
    sub = Subway(url, session)
    for p in prefix.split("/"):
        sub = getattr(sub, p)
    return sub


def get_chain_from_dict(data: Dict[str, Any]) -> Chain:
    stub = get_client()

    # convert to dag and checks
    nodes = []
    edges = []

    # convert to dag and checks
    dag = Dag(**data)
    if not dag.sample:
        raise ValueError("Dag has no sample")
    if not dag.main_in:
        raise ValueError("Dag has no main_in")
    if not dag.main_out:
        raise ValueError("Dag has no main_out")

    # get all the actions by querying the APIs
    dag_nodes = dag.nodes
    actions_map = {}  # this is the map between the cf_id and the node object
    for node in dag_nodes:
        if not node.cf_id and not node.cf_data:
            raise ValueError(f"Action {node.id} has no cf_id or cf_data")
        if node.cf_data:
            # programmatic ones should always be picked from the registry also FE will always send this
            # so server should always check for programatic ones via registry
            if node.cf_data.type == Node.types.PROGRAMATIC:
                try:
                    cf_action = programatic_actions_registry.get(node.cf_id)
                except ValueError:
                    raise ValueError(f"Action {node.id} not found")
            else:
                cf_action = Node.from_dict(node.cf_data.node)
        else:
            cf_action = actions_map.get(node.cf_id, None)

        # check if this action is in the registry
        if not cf_action:
            # check if present in the AI registry
            try:
                # print("ai_actions_registry")
                cf_action = ai_actions_registry.get(node.cf_id)
            except ValueError:
                pass
        if not cf_action:
            # check if present in the programatic registry
            try:
                # print("programatic_actions_registry")
                cf_action = programatic_actions_registry.get(node.cf_id)
            except ValueError:
                pass
        if not cf_action:
            # check available on the API
            try:
                # print("stub.fury.actions.u(node.cf_id)()")
                action, err = stub.fury.actions.u(node.cf_id)()
                if err:
                    raise ValueError(f"Action {node.cf_id} not loaded: {action}")
                cf_action = Node.from_dict(action)
                actions_map[node.cf_id] = cf_action  # cache it
            except:
                raise ValueError(f"Action {node.cf_id} not found")

        # standardsize everything to node
        if not isinstance(cf_action, Node):
            cf_action = Node.from_dict(cf_action)
        cf_action.id = node.id  # override the id
        nodes.append(cf_action)

    # now create all the edges
    dag_edges = dag.edges
    for edge in dag_edges:
        if not (edge.source and edge.target and edge.sourceHandle and edge.targetHandle):
            raise ValueError(f"Invalid edge {edge}")
        edges.append(Edge.from_dict(edge.dict()))

    return Chain(
        nodes=nodes,
        edges=edges,
        sample=dag.sample,
        main_in=dag.main_in,
        main_out=dag.main_out,
    )


def get_chain_from_id(id: str) -> Chain:
    """Helper function to load a chain from the given chatbot ID. This assumed that everything was created correctly

    Example:
        >>> from chainfury.client import get_chain_from_id
        >>> chain = get_chain_from_id("l6lnksln")
        >>> chain

    Args:
        id (str): The id of the chain to load

    Returns:
        Chain: The chain object
    """
    # first we call the API to get the chains
    stub = get_client()
    chain, err = stub.chatbot.u(id)()
    if err:
        raise ValueError(f"Could not get chain with id {id}: {chain}")
    out = get_chain_from_dict(chain["dag"])
    return out


def create_new_chain(name: str, chain: Chain) -> Dict[str, Any]:
    """
    Creates a new chain with the given name and chain. If create_actions is True, it will also create the actions
    on the API.

    Example:
        >>> from chainfury import Chain
        >>> from chainfury.client import create_new_chain
        >>> api_resp = create_new_chain("my chain", chain)
        >>> chain_new = Chain.from_dict(api_resp)
        >>> chain_new

    Args:
        name (str): The name of the chain
        chain (Chain): The chain object
        create_actions (bool, optional): Whether to create the actions or not. Defaults to False.

    Returns:
        Dict[str, Any]: The API response
    """
    stub = get_client()
    chain_dict = chain.to_dict()
    nodes = chain_dict["nodes"]
    dag_nodes = []
    for i, node in enumerate(chain.nodes.values()):
        _node_data = FENode(
            id=node.id,
            position=FENode.Position(x=i * 100, y=i * 100),
            type="FuryEngineNode",
            width=100,
            height=100,
            selected=False,
            position_absolute=FENode.Position(x=i * 100, y=i * 100),
            dragging=False,
        )

        # out, err = stub.fury.actions(
        #     "post",
        #     trailing="/",
        #     json={
        #         "name": node["name"],
        #         "description": "",
        #         "fn": {
        #             "model_id": node["fn"]["model"]["id"],
        #             "model_params": node["fn"]["model_params"],
        #             "fn": node["fn"]["fn"],
        #         },
        #         "outputs": node["outputs"],
        #     },
        # )
        # if err:
        #     raise ValueError(f"Could not create node: {out}")
        # logger.info(f"Created new action with ID: {out['id']}")
        # _node_data.cf_id = out["id"]

        # in this case the entire action is stored with the DAG object
        _node_data.cf_id = node.id
        _node_data.cf_data = FENode.CFData(
            id=node.id,
            type=node.type,
            node=node.to_dict(),
            value=None,
        )

        # add the node to the list
        dag_nodes.append(_node_data.dict())
    chain_dict["nodes"] = dag_nodes

    # update the chain_dict
    edges = []
    for e in chain.edges:
        edges.append(
            EdgeType(
                id=f"{e.src_node_id}/{e.src_node_var}-{e.trg_node_id}/{e.trg_node_var}",
                source=e.src_node_id,
                sourceHandle=e.src_node_var,
                target=e.trg_node_id,
                targetHandle=e.trg_node_var,
            ).dict()
        )
    chain_dict["edges"] = edges
    data = {"name": name, "dag": chain_dict, "engine": "fury"}
    out, err = stub.chatbot("post", trailing="/", json=data)
    if err:
        raise ValueError(f"Could not create chain: {out}")
    return out



from chainfury import ai_actions_registry, Model


# this is a super generic function to reply funnily to something
def hello_world(message: str):
    return {
        "prompt": f"Give a witty response for this: 'Hello dear AI agent, {message}'\n\nWitty Response:",
    }


ai_actions_registry.register(
    node_id="hello-world",
    description="Python function loaded from a file used as an AI action",
    model_id="openai-completion",
    model_params={
        "model": "text-curie-001",
        "temperature": 0.1,
    },
    fn=hello_world,
    outputs={
        "generations": ("choices", 0, "text"),
    },
)


ai_actions_registry.register(
    node_id="deep-rap-quote",
    description="J-type action will write a deep poem in the style of a character",
    model_id="openai-chat",
    model_params={
        "model": "gpt-3.5-turbo",
    },
    fn={
        "messages": [
            {
                "role": "user",
                "content": "give a deep 8 line rap quote on life in the style of {{ character }}.",
            },
        ],
    },
    outputs={
        "chat_reply": ("choices", 0, "message", "content"),
    },
)



import copy
import json
import inspect
import datetime
import traceback
from pprint import pformat
from typing import Any, Union, Optional, Dict, List, Tuple, Callable, Generator
from collections import deque, defaultdict

import jinja2schema
from jinja2schema import model as j2sm

from chainfury.utils import logger, terminal_top_with_text
from chainfury.types import FENode


class Secret(str):
    """This class just means that in Var it will be taken as a password field"""

    def __init__(self, value):
        self.value = value


#
# Vars: this is the base class for all the fields that the user can provide from the front end
#


class Var:
    def __init__(
        self,
        type: Union[str, List["Var"]],
        format: str = "",
        items: List["Var"] = [],
        additionalProperties: Union[List["Var"], "Var"] = [],
        password: bool = False,
        #
        required: bool = False,
        placeholder: str = "",
        show: bool = False,
        name: str = "",
        *,
        loc: Optional[Tuple] = (),
    ):
        """`Var` is a single input / output for a node.

        Args:
            type (Union[str, List[Var]]): The type of the variable. If it is a list, then it is a list of Var objects.
            format (str, optional): The format of the variable. Defaults to "".
            items (List[Var], optional): If the type is a list, then this is the list of Var objects that are in the list. Defaults to [].
            additionalProperties (Union[List[Var], Var], optional): If the type is an object, then this is the list of Var objects that are in the object. Defaults to [].
            password (bool, optional): If the type is a string, then this is whether it is a password field. Defaults to False.
            required (bool, optional): Whether this field is required. Defaults to False.
            placeholder (str, optional): The placeholder text for this field. Defaults to "".
            show (bool, optional): Whether this field should be shown. Defaults to False.
            name (str, optional): The name of this field. Defaults to "".
            loc (Optional[Tuple], optional): The location of this field. Defaults to ().
        """
        self.type = type
        self.format = format
        self.items = items or []
        self.additionalProperties = additionalProperties
        self.password = password
        #
        self.required = required
        self.placeholder = placeholder
        self.show = show
        self.name = name
        #
        self.value = None
        self.loc = loc  # this is the location from which this value is extracted

    def __repr__(self) -> str:
        return f"Var({'*' if self.required else ''}name='{self.name}', type='{self.type}', items={self.items}, additionalProperties={self.additionalProperties})"

    def to_dict(self) -> Dict[str, Any]:
        """Serialise this Var to a dictionary that can be JSON serialised and sent to the client.

        Returns:
            Dict[str, Any]: The serialised Var.
        """
        d: Dict[str, Any] = {"type": self.type}
        if type(self.type) == list and len(self.type) and type(self.type[0]) == Var:
            d["type"] = [x.to_dict() for x in self.type]
        if self.format:
            d["format"] = self.format
        if self.items:
            d["items"] = [item.to_dict() for item in self.items]
        if self.additionalProperties:
            if isinstance(self.additionalProperties, Var):
                d["additionalProperties"] = self.additionalProperties.to_dict()
            else:
                d["additionalProperties"] = self.additionalProperties
        if self.password:
            d["password"] = self.password
        #
        if self.required:
            d["required"] = self.required
        if self.placeholder:
            d["placeholder"] = self.placeholder
        if self.show:
            d["show"] = self.show
        if self.name:
            d["name"] = self.name
        if self.loc:
            d["loc"] = self.loc
        return d

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> "Var":
        """Deserialise a Var from a dictionary.

        Args:
            d (Dict[str, Any]): The dictionary to deserialise from.

        Returns:
            Var: The deserialised Var.
        """
        type_val = d.get("type")
        format_val = d.get("format", "")
        items_val = d.get("items", [])
        additional_properties_val = d.get("additionalProperties", [])
        password_val = d.get("password", False)
        required_val = d.get("required", False)
        placeholder_val = d.get("placeholder", "")
        show_val = d.get("show", False)
        name_val = d.get("name", "")
        loc_val = d.get("loc", ())

        if isinstance(type_val, list):
            type_val = [Var.from_dict(x) if isinstance(x, dict) else x for x in type_val]
        elif isinstance(type_val, dict):
            type_val = Var.from_dict(type_val)

        items_val = [Var.from_dict(x) if isinstance(x, dict) else x for x in items_val]
        additional_properties_val = (
            Var.from_dict(additional_properties_val) if isinstance(additional_properties_val, dict) else additional_properties_val
        )

        var = cls(
            type=type_val,  # type: ignore
            format=format_val,
            items=items_val,
            additionalProperties=additional_properties_val,
            password=password_val,
            required=required_val,
            placeholder=placeholder_val,
            show=show_val,
            name=name_val,
            loc=loc_val,
        )
        return var

    def set_value(self, v: Any):
        """Set the value of this Var.

        Args:
            v (Any): The value to set.
        """
        self.value = v


def pyannotation_to_json_schema(
    x: Any,
    allow_any: bool,
    allow_exc: bool,
    allow_none: bool,
    *,
    trace: bool = False,
) -> Var:
    """Function to convert the given annotation from python to a Var which can then be JSON serialised and sent to the
    clients.

    Args:
        x (Any): The annotation to convert.
        allow_any (bool): Whether to allow the `Any` type.
        allow_exc (bool): Whether to allow the `Exception` type.
        allow_none (bool): Whether to allow the `None` type.
        trace (bool, optional): Adds verbosity the schema generation also set FURY_LOG_LEVEL='debug'. Defaults to False.

    Returns:
        Var: The converted annotation.
    """
    if isinstance(x, type):
        if trace:
            logger.debug("t0")

        if x == str:
            return Var(type="string")
        elif x == int or x == float:
            return Var(type="number")
        elif x == bool:
            return Var(type="boolean")
        elif x == bytes:
            return Var(type="string", format="byte")
        elif x == list:
            return Var(type="array", items=[Var(type="string")])
        elif x == dict:
            return Var(type="object", additionalProperties=Var(type="string"))

        # there are some types that are unique to the fury system
        elif x == Secret:
            return Var(type="string", password=True)
        elif x == Model:
            return Var(type=Model.TYPE_NAME, required=False, show=False)
        if x == Exception and allow_exc:
            return Var(type="exception", required=False, show=False)
        elif x == type(None) and allow_none:
            return Var(type="null", required=False, show=False)
        else:
            raise ValueError(f"i0: Unsupported type: {x}")
    elif isinstance(x, str):
        if trace:
            logger.debug("t1")
        return Var(type="string")
    elif hasattr(x, "__origin__") and hasattr(x, "__args__"):
        if trace:
            logger.debug("t2")
        if x.__origin__ == list:
            if trace:
                logger.debug("t2.1")
            return Var(
                type="array",
                items=[pyannotation_to_json_schema(x=x.__args__[0], allow_any=allow_any, allow_exc=allow_exc, allow_none=allow_none)],
            )
        elif x.__origin__ == dict:
            if len(x.__args__) == 2 and x.__args__[0] == str:
                if trace:
                    logger.debug("t2.2")
                return Var(
                    type="object",
                    additionalProperties=pyannotation_to_json_schema(
                        x=x.__args__[1], allow_any=allow_any, allow_exc=allow_exc, allow_none=allow_none
                    ),
                )
            else:
                raise ValueError(f"i2: Unsupported type: {x}")
        elif x.__origin__ == tuple:
            if trace:
                logger.debug("t2.3")
            return Var(
                type="array",
                items=[
                    pyannotation_to_json_schema(x=arg, allow_any=allow_any, allow_exc=allow_exc, allow_none=allow_none)
                    for arg in x.__args__
                ],
            )
        elif x.__origin__ == Union:
            # Unwrap union types with None type
            types = [arg for arg in x.__args__ if arg is not None]
            if len(types) == 1:
                if trace:
                    logger.debug("t2.4")
                return pyannotation_to_json_schema(x=types[0], allow_any=allow_any, allow_exc=allow_exc, allow_none=allow_none)
            else:
                if trace:
                    logger.debug("t2.5")
                return Var(
                    type=[
                        pyannotation_to_json_schema(x=typ, allow_any=allow_any, allow_exc=allow_exc, allow_none=allow_none) for typ in types
                    ]
                )
        else:
            raise ValueError(f"i3: Unsupported type: {x}")
    elif isinstance(x, tuple):
        if trace:
            logger.debug("t4")
        return Var(
            type="array",
            items=[
                Var(type="string"),
                pyannotation_to_json_schema(x=x[1], allow_any=allow_any, allow_exc=allow_exc, allow_none=allow_none),
            ]
            * len(x),
        )
    elif x == Any and allow_any:
        if trace:
            logger.debug("t5")
        return Var(type="string")
    else:
        if trace:
            logger.debug("t6")
        raise ValueError(f"i4: Unsupported type: {x}")


def func_to_vars(func: object) -> List[Var]:
    """
    Extracts the signature of a function and converts it to an array of Var objects.

    Args:
        func (Callable): The function to extract the signature from.

    Returns:
        List[Var]: The array of Var objects.
    """
    signature = inspect.signature(func)  # type: ignore
    fields = []
    for param in signature.parameters.values():
        schema = pyannotation_to_json_schema(param.annotation, allow_any=False, allow_exc=False, allow_none=False)
        schema.required = param.default is inspect.Parameter.empty
        schema.name = param.name
        schema.placeholder = str(param.default) if param.default is not inspect.Parameter.empty else ""
        if not schema.name.startswith("_"):
            schema.show = True
        fields.append(schema)
    return fields


def func_to_return_vars(func, returns: Dict[str, Tuple]) -> List[Var]:
    """
    Analyses the return annotation type of the signature of a function and converts it to an array of named Var objects.

    Args:
        func (Callable): The function to extract the signature from.
        returns (Dict[str, Tuple]): The dictionary of return types.

    Returns:
        List[Var]: The array of Var objects.
    """
    signature = inspect.signature(func)
    schema = pyannotation_to_json_schema(signature.return_annotation, allow_any=False, allow_exc=True, allow_none=True)
    if not (
        schema.type == "array"
        and len(schema.items) == 2
        and type(schema.items[1].type) == list
        and any(x.type == "exception" for x in schema.items[1].type)
    ):
        raise ValueError("Interface requires return type Tuple[..., Optional[Exception]] where ... is JSON serializable")

    # take the names provided in returns and populate the returning field
    logger.debug(f"RETURNS: {returns}")
    ret = schema.items[0]
    logger.debug(f"RET: {ret}")
    if ret.type == "array":
        assert len(returns) in [1, len(ret.items)], f"For array outputs, returns should either be 1 or {len(ret.items)}, got {len(returns)}"
        if len(returns) == 1:
            ret.items[0].name = next(iter(returns))
            ret.items[0].loc = returns[next(iter(returns))]
        for i, n in zip(ret.items, returns):
            i.name = n
            i.loc = returns[n]
        ret = ret.items
    else:
        assert len(returns) == 1, "Items that are not arrays can have only 1 returning var. This can also be a bug"
        ret.name = next(iter(returns))
        ret.loc = returns[next(iter(returns))]
        ret = [
            ret,
        ]
    logger.debug(f"FINAL: {ret}")
    return ret


def jinja_schema_to_vars(v) -> Var:
    """
    Converts a Jinja schema to a Var object.

    Args:
        v ([type]): The Jinja schema.

    Returns:
        Var: The Var object.
    """
    if type(v) == j2sm.Scalar or type(v) == j2sm.String:
        field = Var(type="string", required=True)
    elif type(v) == j2sm.Number:
        field = Var(type="number", required=True)
    elif type(v) == j2sm.Boolean:
        field = Var(type="boolean", required=True)
    elif type(v) == j2sm.Unknown:
        field = Var(type="string", required=True)
    elif type(v) == j2sm.Variable:
        field = Var(type="string", required=True)
    elif type(v) == j2sm.Dictionary:
        field = Var(type="object", required=True)
        all_fields = []
        for k, v in v.items():
            field_item = jinja_schema_to_vars(v)
            field_item.name = k
            all_fields.append(field_item)
        field.additionalProperties = all_fields
    elif type(v) == j2sm.List:
        field = Var(type="array", required=True)
        field.items = [jinja_schema_to_vars(v.item)]
    elif type(v) == j2sm.Tuple:
        field = Var(type="array", required=True)
        if v.items:
            field.items = [jinja_schema_to_vars(x) for x in v.items]
    else:
        raise ValueError(f"cannot handle type {type(v)}")
    return field


def jtype_to_vars(prompt: str) -> List[Var]:
    """
    Converts a Jinja prompt to an array of Var objects.

    Args:
        prompt (str): The Jinja prompt.

    Returns:
        List[Var]: The array of Var objects.
    """
    try:
        s = jinja2schema.infer(prompt)
        fields = []
        for k, v in s.items():
            f = jinja_schema_to_vars(v)
            f.name = k
            fields.append(f)
    except Exception as e:
        logger.error(
            "Could not parse prompt to jinja schema. We support only for/if/filters in jinja2. "
            "Please read here for more information: https://jinja.palletsprojects.com/en/3.1.x/templates/"
        )
        raise e
    return fields


def extract_jinja_indices(data: Union[str, List, Dict[str, Any]], current_index=(), indices=None) -> List:
    """
    This takes in a nested object and returns all the locations where jinja template was detected.

    Args:
        data (Union[str, List, Dict[str, Any]]): The nested object.
        current_index (tuple, optional): The current index. Defaults to ().
        indices ([type], optional): The indices. Defaults to None.

    Returns:
        List: The list of indices.

    Example:
        >>> from chainfury.base import extract_jinja_indices
        >>> extract_jinja_indices({
        ...     "foo": {
        ...         "bar": "{{ baz }}",
        ...         "gaa": ["{{ joo }}"]
        ...    }
        ... })
        [(('foo', 'bar'), Var(...)), (('foo', 'gaa', '0'), Var(...))]
        >>> # more examples
        [(('3', 'content'), [Var('num1', type=string, items=[], additionalProperties=[]), Var('num2', type=string, items=[], additionalProperties=[])])]
        [((), [Var('message', type=string, items=[], additionalProperties=[])])]
        [(('meta_prompt', 'data'), [Var('place', type=string, items=[], additionalProperties=[])])]
        [('0', [Var('name', type=string, items=[], additionalProperties=[])])]
        [(('meta', 'ptype'), [Var('genome', type=string, items=[], additionalProperties=[])])]
        [(('level-0', 'level-1', 'level-2'), [Var('thing', type=string, items=[], additionalProperties=[])]), (('level-0', 'nice'), [Var('feeling', type=string, items=[], additionalProperties=[])])]
        []
    """
    if indices is None:
        indices = []

    if isinstance(data, str):
        fields = jtype_to_vars(data)
        if fields:
            indices.append((current_index, fields))
    elif isinstance(data, list):
        for i, item in enumerate(data):
            if current_index:
                if type(current_index) == tuple:
                    new_index = (*current_index, i)
                else:
                    new_index = (current_index, i)
            else:
                new_index = str(i)
            extract_jinja_indices(data=item, current_index=new_index, indices=indices)
    elif isinstance(data, dict):
        for key, value in data.items():
            if current_index:
                if type(current_index) == tuple:
                    new_index = (*current_index, key)
                else:
                    new_index = (current_index, key)
            else:
                new_index = key
            extract_jinja_indices(data=value, current_index=new_index, indices=indices)

    return indices


def get_value_by_keys(obj, keys, *, _first_sentinal: bool = False) -> Any:
    """Takes in an arbitrary nested object and returns the value at the location specified by the keys.

    Args:
        obj (Union[List, Dict[str, Any]]): The nested object.
        keys (Union[str, List[str], Tuple[str, ...]]): The keys. See `extract_jinja_indices` for examples.
        _first_sentinal (bool, optional): flag to tell if this is the first input or not, user should not use this. Defaults to False.

    Returns:
        Any: The value at the location specified by the keys.
    """
    if not keys:
        return obj

    keys = (keys,) if not isinstance(keys, (list, tuple)) else keys
    key = keys[0]

    if key == "*":
        if not _first_sentinal:
            raise ValueError("gvk1: Cannot use wildcard '*' as first key")

        # If the key is "*", apply the subsequent keys to all elements in the current list or dictionary.
        if isinstance(obj, list):
            return [get_value_by_keys(elem, keys[1:], _first_sentinal=True) for elem in obj]
        elif isinstance(obj, dict):
            return {k: get_value_by_keys(v, keys[1:], _first_sentinal=True) for k, v in obj.items()}

    if isinstance(obj, dict):
        return get_value_by_keys(obj.get(key), keys[1:], _first_sentinal=True)
    elif isinstance(obj, (tuple, list)):
        try:
            key = int(key)
        except ValueError:
            raise ValueError(f"gvk2: Cannot use key '{key}' on a list")
        if not type(key) == int:
            raise ValueError(f"gvk3: Cannot use key '{key}' on a list")
        key = int(key)
        if isinstance(key, int) and 0 <= key < len(obj):
            return get_value_by_keys(obj[key], keys[1:], _first_sentinal=True)

    return None


def put_value_by_keys(obj, keys, value: Any):
    """Takes in an arbitrary nested object and sets the value at the location specified by the keys.

    Args:
        obj (Union[List, Dict[str, Any]]): The nested object.
        keys (Union[str, List[str], Tuple[str, ...]]): The keys. See `extract_jinja_indices` for examples.
        value (Any): The value to set.
    """
    if not keys:
        return

    keys = (keys,) if not isinstance(keys, (list, tuple)) else keys
    key = keys[0]
    if len(keys) == 1:
        if isinstance(obj, dict):
            obj[key] = value
        elif isinstance(obj, list) and isinstance(key, int) and 0 <= key < len(obj):
            obj[key] = value
    else:
        if isinstance(obj, dict):
            if key not in obj or not isinstance(obj[key], (dict, list)):
                obj[key] = {} if isinstance(keys[1], str) else []
            put_value_by_keys(obj[key], keys[1:], value)
        elif isinstance(obj, list) and isinstance(key, int) and 0 <= key < len(obj):
            if not isinstance(obj[key], (dict, list)):
                obj[key] = {} if isinstance(keys[1], str) else []
            put_value_by_keys(obj[key], keys[1:], value)


#
# Model: Each model is the processing engine of the AI actions. It is responsible for keeping
#        the state of each of the wrapped functions for different API calls.
#


class Model:
    TYPE_NAME = "model"
    """constant for the type name"""

    def __init__(
        self,
        collection_name: str,
        id: str,
        fn: object,
        description,
        usage: List[Union[str, int]] = [],
        tags=[],
    ):
        """Defines a single callable model.

        Args:
            collection_name (str): The name of the collection.
            id (str): The id of the model.
            fn (Callable): The callable to wrap.
            description (str): The description of the model.
            usage (List[Union[str, int]], optional): The location that tells usage for a call. Defaults to [].
            tags (List[str], optional): The tags for the model. Defaults to [].
        """
        self.collection_name = collection_name
        self.id = id
        self.fn = fn
        self.description = description
        self.usage = usage
        self.vars = func_to_vars(fn)
        self.tags = tags

    def __repr__(self) -> str:
        return f"Model('{self.collection_name}', '{self.id}')"

    def to_dict(self, no_vars: bool = False) -> Dict[str, Any]:
        """Converts the model to a dictionary.

        Args:
            no_vars (bool, optional): Whether to include the vars. Defaults to False.

        Returns:
            Dict[str, Any]: The dictionary representation of the model.
        """
        return {
            "collection_name": self.collection_name,
            "id": self.id,
            "description": self.description,
            "usage": self.usage,
            "vars": [x.to_dict() for x in self.vars] if not no_vars else [],
            "tags": self.tags,
        }

    def __call__(self, model_data: Dict[str, Any]) -> Tuple[Any, Optional[Exception]]:
        """Calls the model with the given data.

        Args:
            model_data (Dict[str, Any]): The data to pass to the model.

        Returns:
            Tuple[Any, Optional[Exception]]: The result of the model and the exception if any.
        """
        try:
            out = self.fn(**model_data)  # type: ignore
            return out, None
        except Exception as e:
            return traceback.format_exc(), e


#
# Node: Each box that is drag and dropped in the UI is a Node, it will tell what kind of things are
#       its inputs, outputs and fields that are shown in the UI. It can be of different types and
#       it only wraps teh
#


class NodeType:
    PROGRAMATIC = "programatic"
    """constant for the programatic node type"""
    AI = "ai-powered"
    """constant for the AI node type"""
    MEMORY = "memory"
    """constant for the memory node type"""


class Node:
    types = NodeType()

    def __init__(
        self,
        id: str,
        type: str,
        fn: object,  # the function to call
        fields: List[Var],
        outputs: List[Var],
        description: str = "",
        tags: List[str] = [],
    ):
        """Node is a single unit of computation in a Dag. All the actions are considered as nodes.

        Args:
            id (str): The id of the node.
            type (str): The type of the node. See `Node.types` for valid types.
            fn (object): The function to call.
            fields (List[Var]): The fields of the node.
            outputs (List[Var]): The outputs of the node.
            description (str, optional): The description of the node. Defaults to "".
            tags (List[str], optional): The tags for the node. Defaults to [].
        """
        # some bacic checks
        _valid_types = [getattr(NodeType, x) for x in dir(NodeType) if not x.startswith("__")]
        if type not in _valid_types:
            raise ValueError(f"Invalid node type: {type}, {_valid_types}")

        # set the values
        self.id = id
        self.type = type
        self.description = description
        self.fields = fields
        self.outputs = outputs
        self.fn = fn
        self.tags = tags

    def __repr__(self) -> str:
        out = f"FuryNode{{ ('{self.id}', '{self.type}') ["
        for f in self.fields:
            if f.required:
                out += f"\n      {f},"
        out += f"\n] ({len(self.fields)}) => ({len(self.outputs)}) ["
        for o in self.outputs:
            out += f"\n      {o},"
        out += f"\n] }}"
        return out

    def has_field(self, field: str) -> bool:
        """helper function to check if the node has a field with the given name.

        Args:
            field (str): The name of the field to check.

        Returns:
            bool: True if the node has the field, False otherwise.
        """
        return any([x.name == field for x in self.fields])

    def to_dict(self) -> Dict[str, Any]:
        """Converts the node to a dictionary.

        Returns:
            Dict[str, Any]: The dictionary representation of the node.
        """
        from chainfury.agent import AIAction, Memory

        fn = {}
        name = self.id
        if isinstance(self.fn, AIAction):
            fn = self.fn.to_dict(no_vars=True)
            name = fn.pop("action_name")
        elif isinstance(self.fn, Memory):
            fn = self.fn.to_dict()

        return {
            "id": self.id,
            "type": self.type,
            "fn": fn,
            "name": name,
            "description": self.description,
            "fields": [field.to_dict() for field in self.fields],
            "outputs": [o.to_dict() for o in self.outputs],
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any], verbose: bool = False) -> "Node":
        """Creates a node from a dictionary.

        Args:
            data (Dict[str, Any]): The dictionary representation of the node.
            verbose (bool, optional): Whether to print verbose logs. Defaults to False.

        Returns:
            Node: The node created from the dictionary.
        """
        if verbose:
            logger.info("Creating node from dict: %s", data)
        fields = [Var.from_dict(x) for x in data["fields"]]
        outputs = [Var.from_dict(x) for x in data["outputs"]]
        fn = data["fn"]
        if not fn:
            raise ValueError(f"Invalid fn: {fn}")

        from chainfury.agent import AIAction, Memory

        node_type = data["type"]
        if node_type == NodeType.AI:
            fn = AIAction.from_dict(fn)
        elif node_type == NodeType.MEMORY:
            fn = Memory.from_dict(fn)

        return cls(
            id=data["id"],
            type=node_type,
            fn=fn,
            description=data["description"],
            fields=fields,
            outputs=outputs,
        )

    def to_json(self, indent=None) -> str:
        """Converts the node to a json string.

        Args:
            indent (int, optional): The indent to use. Defaults to None.

        Returns:
            str: The json string representation of the node.
        """
        return json.dumps(self.to_dict(), indent=indent)

    @classmethod
    def from_json(cls, data: str) -> "Node":
        """Creates a node from a json string.

        Args:
            data (str): The json string representation of the node.

        Returns:
            Node: The node created from the json string.
        """
        return cls.from_dict(json.loads(data))

    def __call__(self, data: Dict[str, Any], print_thoughts: bool = False) -> Tuple[Any, Optional[Exception]]:
        """Calls the node with the given data.

        Args:
            data (Dict[str, Any]): The data to pass to the node.
            print_thoughts (bool, optional): Whether to print the thoughts of the node, useful for debugging. Defaults to False.

        Returns:
            Tuple[Any, Optional[Exception]]: The result of the node and the exception if any.
        """
        data_keys = set(data.keys())
        template_keys = set([x.name for x in self.fields])
        try:
            if not data_keys.issubset(template_keys):
                raise ValueError(f"Invalid keys passed to node '{self.id}': {data_keys - template_keys}")
            if print_thoughts:
                print(terminal_top_with_text(f"Node: {self.id}"))
                print("Inputs:\n------")
                print(pformat(data))

            _out = self.fn(**data)  # type: ignore
            out = _out[0] if isinstance(_out, tuple) else _out
            err = _out[1] if isinstance(_out, tuple) and len(_out) > 1 else None
            if err:
                raise err

            # this is where we have to polish this outgoing result into the structure as configured in self.outputs
            logger.debug(f"> fn_out: {out}")
            logger.debug(f"> OUTPUTS: {self.outputs}")
            for o in self.outputs:
                # logger.debug("  OP:", o.name, o._loc)
                o.set_value(get_value_by_keys(out, o.loc))

            fout = {o.name: o.value for o in self.outputs}
            if print_thoughts:
                print("Outputs:\n-------")
                print(pformat(fout))
            return fout, None
        except Exception as e:
            tb = traceback.format_exc()
            return tb, e

    def forward(self, *args, **kwargs):
        """This is a convinience method so users can subclass this and implement their own forward method while the
        underlying processing implemented within __call__ method remain intact."""
        return self(*args, **kwargs)


#
# Edge: Each connection between two boxes on the UI is called an Edge, it is only a dataclass without any methods.
#


class Edge:
    """Creates an edge between two nodes.

    Args:
        src_node_id (str): The id of the source node.
        src_node_var (str): The name of the source node variable.
        trg_node_id (str): The id of the target node.
        trg_node_var (str): The name of the target node variable.
    """

    def __init__(
        self,
        src_node_id: str,
        src_node_var: str,
        trg_node_id: str,
        trg_node_var,
    ):
        self.src_node_id = src_node_id
        self.trg_node_id = trg_node_id
        self.src_node_var = src_node_var
        self.trg_node_var = trg_node_var

    def __repr__(self) -> str:
        out = f"FuryEdge('{self.src_node_id}/{self.src_node_var}' => '{self.trg_node_id}/{self.trg_node_var}')"
        return out

    def to_dict(self) -> Dict[str, Any]:
        """Serializes the edge to a dictionary.

        Returns:
            Dict[str, Any]: The dictionary representation of the edge.
        """
        return {
            "source": self.src_node_id,
            "sourceHandle": self.src_node_var,
            "target": self.trg_node_id,
            "targetHandle": self.trg_node_var,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any], verbose: bool = False) -> "Edge":
        """Creates an edge from a dictionary.

        Args:
            data (Dict[str, Any]): The dictionary representation of the edge.

        Returns:
            Edge: The edge created from the dictionary.
        """
        return cls(
            data["source"],
            data["sourceHandle"],
            data["target"],
            data["targetHandle"],
        )


#
# Dag: An entire flow is called the Chain
#


class Chain:
    """A chain is a full flow of nodes and edges.

    Args:
        nodes (List[Node], optional): The list of nodes in the chain. Defaults to [].
        edges (List[Edge], optional): The list of edges in the chain. Defaults to [].
        sample (Dict[str, Any], optional): The sample data to use for the chain. Defaults to {}.
        main_in (str, optional): The name of the input var for the chat input. Defaults to "".
        main_out (str, optional): The name of the output var for the chat output. Defaults to "".
    """

    def __init__(
        self,
        nodes: List[Node] = [],
        edges: List[Edge] = [],
        *,
        sample: Dict[str, Any] = {},
        main_in: str = "",
        main_out: str = "",
    ):
        self.nodes: Dict[str, Node] = {node.id: node for node in nodes}
        self.edges = edges

        if len(self.nodes) == 1:
            assert len(self.edges) == 0, "Cannot have edges with only 1 node"
            self.topo_order = [next(iter(self.nodes))]
        else:
            self.topo_order = topological_sort(self.edges)
        self.sample = sample
        self.main_in = main_in
        self.main_out = main_out

        for node_id in self.topo_order:
            assert node_id in self.nodes, f"Missing node from an edge: {node_id}"

        # to a dry run to validate everything
        self.to_dict()

    def __repr__(self) -> str:
        # return f"FuryDag(nodes: {len(self.nodes)}, edges: {len(self.edges)})"
        out = "FuryDag(\n  nodes: ["
        for n in self.nodes:
            out += f"\n    {n},"
        out += "\n  ],\n  edges: ["
        for e in self.edges:
            out += f"\n    {e},"
        out += f"\n  ]\n  main_in: {self.main_in}\n  main_out: {self.main_out}\n)"
        return out

    def to_dict(self, main_in: str = "", main_out: str = "", sample: Dict[str, Any] = {}) -> Dict[str, Any]:
        """Serializes the chain to a dictionary.

        Args:
            main_in (str, optional): The name of the input var for the chat input. Defaults to "".
            main_out (str, optional): The name of the output var for the chat output. Defaults to "".
            sample (Dict[str, Any], optional): The sample data to use for the chain. Defaults to {}.

        Returns:
            Dict[str, Any]: The dictionary representation of the chain.
        """
        main_in = main_in or self.main_in
        main_out = main_out or self.main_out
        sample = sample or self.sample
        if main_in not in sample:
            logger.error(f"Key should be present in 'sample': {main_in}")
        # assert main_in in sample, f"Invalid key: {main_in}"

        if not (main_in or main_out or sample):
            logger.warning("No main_in, main_out or sample provided, using defaults")
            raise ValueError("No main_in, main_out or sample provided, using defaults")
        return {
            "nodes": [node.to_dict() for node in self.nodes.values()],
            "edges": [edge.to_dict() for edge in self.edges],
            "topo_order": self.topo_order,
            "sample": sample,
            "main_in": main_in,
            "main_out": main_out,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any], verbose: bool = False) -> "Chain":
        """Creates a chain from a dictionary.

        Args:
            data (Dict[str, Any]): The dictionary representation of the chain.

        Returns:
            Chain: The chain created from the dictionary.
        """
        nodes = [Node.from_dict(data=x, verbose=verbose) for x in data["nodes"]]
        edges = [Edge.from_dict(data=x, verbose=verbose) for x in data["edges"]]
        return cls(nodes=nodes, edges=edges, sample=data["sample"], main_in=data["main_in"], main_out=data["main_out"])

    def to_json(self) -> str:
        """Serializes the chain to a JSON string.

        Returns:
            str: The JSON string representation of the chain.
        """
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, data: str):
        """Creates a chain from a JSON string.

        Args:
            data (str): The JSON string representation of the chain.

        Returns:
            Chain: The chain created from the JSON string.
        """
        return cls.from_dict(json.loads(data))

    @classmethod
    def from_id(cls, id: str):
        from chainfury.client import get_chain_from_id

        return get_chain_from_id(id)

    def step(
        self,
        node_id: str,
        pre_data: Dict[str, Any],
        full_ir: Dict[str, Any],
        print_thoughts: bool = False,
        thoughts_callback: Optional[Callable] = None,
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """Performs a single step in the chain, useful for manual debugging.

        Args:
            node_id (str): The id of the node to step.
            pre_data (Dict[str, Any]): The data to use for the step.
            full_ir (Dict[str, Any]): The full IR to use for the step.
            print_thoughts (bool, optional): Whether to print the thoughts. Defaults to False.
            thoughts_callback (Optional[Callable], optional): A callback to call with the thoughts. Defaults to None.

        Returns:
            Tuple[Dict[str, Any], Dict[str, Any]]: The currrent output and updated thoughts ir buffer.
        """
        node = self.nodes[node_id]
        incoming_edges = list(filter(lambda edge: edge.trg_node_id == node_id, self.edges))

        # clear out all the nodes that this thing needs into a separate rep
        logger.debug(f"Processing node: {node_id}")
        logger.debug(f"Current full_ir: {set(full_ir.keys())}")
        _data = {}

        # first check if this node has any fields that are in the data
        all_keys = list(pre_data.keys())
        for k in all_keys:
            if node.has_field(k):
                _data[k] = pre_data[k]  # don't pop this, some things are shared between actions eg. openai_api_key
            elif k.startswith(node.id):
                _data[k.split("/", 1)[1]] = pre_data.pop(k)  # pop this, it is not needed anymore

        # then merge from the ir buffer
        for edge in incoming_edges:
            logger.debug(f"Incoming edge: {edge}")
            req_key = f"{edge.src_node_id}/{edge.src_node_var}"
            logger.debug(f"Looking for key: {req_key}")
            # need to check if this information is available in the IR buffer, if it is not then this is an error
            ir_value = pre_data.get(req_key, None) or full_ir.get(req_key, {}).get("value", None)
            if ir_value is None:
                raise ValueError(f"Missing value for {req_key}")
            _data[edge.trg_node_var] = ir_value

        # then run the node
        out, err = node.forward(_data, print_thoughts=print_thoughts)
        if err:
            logger.error(f"TRACE: {out}")
            raise err

        yield_dict = {}
        for k, v in out.items():
            key = f"{node_id}/{k}"
            value = {
                "value": v,
                "timestamp": datetime.datetime.now().isoformat(),
            }
            full_ir[key] = value
            thought = {"key": key, **value}
            yield_dict[key] = value
            if thoughts_callback is not None:
                thoughts_callback(thought)
                if print_thoughts:
                    print(thought)

        return yield_dict, full_ir

    def __call__(
        self,
        data: Union[str, Dict[str, Any]],
        thoughts_callback: Optional[Callable] = None,
        print_thoughts: bool = False,
    ) -> Tuple[Var, Dict[str, Any]]:
        """
        Runs the chain on the given data. In this function it will run a full dataflow engine along with thoughts buffer
        and a simple callback system at each step.

        Example:
            >>> chain = Chain(...)
            >>> out, thoughts = chain("Hello world")
            >>> print(out)
            ... The first man chuckled and shook his head, "You always have the weirdest explanations for everything."
            >>> print(thoughts)
            ... {
                    '38c813a2-850c-448b-8cfb-bd5775cc4b61/answer': {
                        'timestamp': '2023-06-27T16:50:04.178833',
                        'value': '...'
                    }
                    '1378538b-a15e-475b-9a9d-a31a261165c0/out': {
                        'timestamp': '2023-06-27T16:50:07.818709',
                        'value': '...'
                    }
                }

        You can also stream the intermediate responses by setting using `stream_call` method. You can get the exact same
        result as above by iterating over the response and getting the last response.

        Args:
            data (Union[str, Dict[str, Any]]): The data to run the chain on.
            thoughts_callback (Optional[Callable], optional): The callback function to call at each step. Defaults to None.
            print_thoughts (bool, optional): Whether to print the thoughts buffer at each step. Defaults to False.
            stream (bool, optional): Whether to stream the output or not. Defaults to False.

        Returns:
            Tuple[Var, Dict[str, Any]]: The output of the chain and the thoughts buffer.
        """
        if not isinstance(data, dict):
            assert isinstance(data, str), f"Invalid data type: {type(data)}"
            assert self.sample and self.main_in, "Cannot run a chain without a sample and main_in for string input, please use a dict input"
            data = {self.main_in: data}
        _data = copy.deepcopy(self.sample)  # don't corrupt yourself over multiple calls
        _data.update(data)
        data = _data

        if print_thoughts:
            print(terminal_top_with_text("Chain Starts"))
            print("Inputs:\n------")
            print(pformat(data))

        full_ir = {}
        out = None
        for node_id in self.topo_order:
            yield_dict, full_ir = self.step(
                node_id=node_id,
                pre_data=data,
                full_ir=full_ir,
                print_thoughts=print_thoughts,
                thoughts_callback=thoughts_callback,
            )

        if self.main_out:
            out = full_ir.get(self.main_out)["value"]  # type: ignore

        if print_thoughts:
            print(terminal_top_with_text("Chain Last"))
            print("Outputs:\n------")
            print(pformat(out))
            print(terminal_top_with_text("Chain Ends"))

        return out, full_ir  # type: ignore

    def stream(
        self,
        data: Union[str, Dict[str, Any]],
        thoughts_callback: Optional[Callable] = None,
        print_thoughts: bool = False,
    ) -> Generator[Tuple[Union[Any, Dict[str, Any]], bool], None, None]:
        """
        This is a streaming version of __call__ method. It will yield the intermediate responses as they come in.

        Example:
            >>> chain = Chain(...)
            >>> cf_response_gen = chain.stream("Hello world")
            >>> out = None
            >>> thoughts = {}
            >>> for ir, done in cf_response_gen:
            ...     if done:
            ...         out = ir
            ...     else:
            ...         thoughts.update(ir)
            >>> print(out)
            ... The first man chuckled and shook his head, "You always have the weirdest explanations for everything."
            >>> print(thoughts)
            ... {
                    '38c813a2-850c-448b-8cfb-bd5775cc4b61/answer': {
                        'timestamp': '2023-06-27T16:50:04.178833',
                        'value': '...'
                    }
                    '1378538b-a15e-475b-9a9d-a31a261165c0/out': {
                        'timestamp': '2023-06-27T16:50:07.818709',
                        'value': '...'
                    }
                }

        Args:
            data (Union[str, Dict[str, Any]]): The data to run the chain on.
            thoughts_callback (Optional[Callable], optional): The callback function to call at each step. Defaults to None.
            print_thoughts (bool, optional): Whether to print the thoughts buffer at each step. Defaults to False.

        Yields:
            Generator[Tuple[Union[Any, Dict[str, Any]], bool], None, None]: The intermediate responses and whether the
            response is the final response or not.
        """
        if not isinstance(data, dict):
            assert isinstance(data, str), f"Invalid data type: {type(data)}"
            assert self.sample and self.main_in, "Cannot run a chain without a sample and main_in for string input, please use a dict input"
            data = {self.main_in: data}
        _data = copy.deepcopy(self.sample)  # don't corrupt yourself over multiple calls
        _data.update(data)
        data = _data

        if print_thoughts:
            print(terminal_top_with_text("Chain Starts"))
            print("Inputs:\n------")
            print(pformat(data))

        full_ir = {}
        out = None
        for node_id in self.topo_order:
            yield_dict, full_ir = self.step(
                node_id=node_id,
                pre_data=data,
                full_ir=full_ir,
                print_thoughts=print_thoughts,
                thoughts_callback=thoughts_callback,
            )
            yield yield_dict, False

        if print_thoughts:
            print(terminal_top_with_text("Chain Last"))
            print("Outputs:\n------")
            print(pformat(out))
            print(terminal_top_with_text("Chain Ends"))

        if self.main_out:
            out = full_ir.get(self.main_out)["value"]  # type: ignore
        yield out, True


#
# helper functions
#


class NotDAGError(Exception):
    pass


def edge_array_to_adjacency_list(edges: List[Edge]):
    adjacency_lists = {}
    for edge in edges:
        src = edge.src_node_id
        dst = edge.trg_node_id
        if src not in adjacency_lists:
            adjacency_lists[src] = []
        adjacency_lists[src].append(dst)
    return adjacency_lists


def adjacency_list_to_edge_map(adjacency_list) -> List[Edge]:
    edges = []
    for src, dsts in adjacency_list.items():
        for dst in dsts:
            edges.append(Edge(src_node_id=src, src_node_var="", trg_node_id=dst, trg_node_var=""))
    return edges


def topological_sort(edges: List[Edge]) -> List[str]:
    """Topological sort of a DAG, raises NotDAGError if the graph is not a DAG. This is full proof version
    which will work even if the DAG contains several unconnected chains.

    Args:
        edges (List[Edge]): The edges of the DAG

    Returns:
        List[str]: The topologically sorted list of node ids
    """
    adjacency_lists = edge_array_to_adjacency_list(edges)
    in_degree = defaultdict(int)
    for src, dsts in adjacency_lists.items():
        for dst in dsts:
            in_degree[dst] += 1

    # Add all nodes with no incoming edges to the queue
    queue = deque()
    for node in adjacency_lists:
        if in_degree[node] == 0:
            queue.append(node)

    # For each node, remove it from the graph and add it to the sorted list
    sorted_list = []
    edge_nodes_cntr = 0
    while queue:
        node = queue.popleft()
        sorted_list.append(node)
        neighbours = adjacency_lists.get(node, [])
        if not neighbours:
            edge_nodes_cntr += 1
        for neighbor in neighbours:
            in_degree[neighbor] -= 1
            if in_degree[neighbor] == 0:
                queue.append(neighbor)

    # Check to see if all edges are removed
    if len(sorted_list) == len(adjacency_lists) + edge_nodes_cntr:
        return sorted_list
    else:
        raise NotDAGError("A cycle exists in the graph.")



"""
Engineering Notes
-----------------

You need to have `stability_sdk` installed to use this component. You can install it with:

.. code-block:: bash
    
    pip install chainfury[stability]
    # or to install all the components, note this will keep on growing
    pip install chainfury[all]
"""
import os
from typing import List, Dict, Union, Tuple, Optional

from chainfury import Secret, model_registry, Model, exponential_backoff, DoNotRetryException
from chainfury.utils import CFEnv, store_blob
from chainfury.components.const import Env

try:
    from grpc import RpcError, StatusCode
    from stability_sdk.client import StabilityInference, process_artifacts_from_answers
    from stability_sdk.utils import sampler_from_string
    from stability_sdk.interfaces.gooseai.generation.generation_pb2 import Artifact
    from stability_sdk.interfaces.gooseai.generation.generation_pb2 import ARTIFACT_IMAGE

    STABILITY_SDK_INSTALLED = True
except:
    STABILITY_SDK_INSTALLED = False


def stability_text_to_image(
    text_prompts: Union[str, List[str], List[Dict[str, Union[str, float]]]],
    stability_api_key: Secret = Secret(""),
    stability_host: str = "grpc.stability.ai:443",
    height: int = 512,
    width: int = 512,
    cfg_scale: float = 7,
    engine: str = "stable-diffusion-v1-5",  # stable-diffusion-xl-1024-v0-9
    start_schedule: float = 1.0,
    end_schedule: float = 0.01,
    sampler: str = "",
    samples: int = 1,
    seed: int = 0,
    steps: int = 50,
    style_preset: str = "",
    *,
    prefix: str = "nbx-cf-component-stability-text-to-image",
    max_retries: int = 3,
    retry_delay: int = 1,
) -> List[str]:
    """
    Generate an image from text prompts using the Stability API.

    Args:
        text_prompts (List[Dict[str, Union[str, float]]]): An array of text prompts to use for generation. Given a text prompt with the text "A lighthouse on a cliff" and a weight of 0.5, it would be represented as:
            [{"text": "A lighthouse on a cliff", "weight": 0.5}]
        stability_api_key (Secret): Your Stability API key. If not provided, we'll try to use the STABILITY_KEY environment variable.
        stability_host (str): The Stability API host to use. Default: "grpc.stability.ai:443".
        height (int): Height of the image in pixels. Must be in increments of 64 and pass the following validation:
            For 768 engines: 589,824 ≤ height * width ≤ 1,048,576
            All other engines: 262,144 ≤ height * width ≤ 1,048,576
        width (int): Width of the image in pixels. Must be in increments of 64 and pass the following validation:
            For 768 engines: 589,824 ≤ height * width ≤ 1,048,576
            All other engines: 262,144 ≤ height * width ≤ 1,048,576
        cfg_scale (float): How strictly the diffusion process adheres to the prompt text (higher values keep your image closer to your prompt). Default: 7.
        sampler (str): Which sampler to use for the diffusion process. If this value is omitted we'll automatically select an appropriate sampler for you.
        samples (int): Number of images to generate. Default: 1.
        seed (int): Random noise seed (omit this option or use 0 for a random seed). Default: 0.
        steps (int): Number of diffusion steps to run. Default: 50.
        style_preset (str): Pass in a style preset to guide the image model towards a particular style. This list of style presets is subject to change.
        prefix (str): Prefix to use for the generated images. Default: "nbx-cf-component-stability-text-to-image/".

    Returns:
        List[str]: A list of base64-encoded PNG images.
    """
    # perform checks
    if not STABILITY_SDK_INSTALLED:
        raise Exception(
            "stability_sdk is not installed, cannot use with chainfury. Please install it with `pip install chainfury[stability]`"
        )
    if not stability_api_key:
        stability_api_key = Secret(Env.STABILITY_KEY("")).value
    if not stability_api_key:
        raise Exception("Stability API key not found. Please set STABILITY_KEY environment variable or pass through function")
    if isinstance(text_prompts, (list, tuple)):
        if len(text_prompts) != 1:
            raise Exception("Only one text prompt is supported at this time")
        prompt = text_prompts[0]
        if isinstance(prompt, dict):
            prompt = prompt["text"]
    else:
        if not isinstance(text_prompts, str):
            raise Exception("text_prompts must be a string or a list of strings")
        prompt = text_prompts

    # make request
    request = {}
    if sampler:
        request["sampler"] = sampler_from_string(sampler)
    if steps:
        request["steps"] = steps
    stability_api = StabilityInference(
        host=stability_host,
        key=stability_api_key,
        engine=engine,
        verbose=True,
    )

    def _fn():
        try:
            answers = stability_api.generate(
                prompt=prompt,  # type: ignore
                height=height,
                width=width,
                start_schedule=start_schedule,
                end_schedule=end_schedule,
                cfg_scale=cfg_scale,
                seed=seed,
                samples=samples,
                style_preset=style_preset,
                **request,
            )
        except RpcError as e:
            if e.code() in [  # type: ignore
                StatusCode.INVALID_ARGUMENT,
                StatusCode.UNAUTHENTICATED,
                StatusCode.UNIMPLEMENTED,
                StatusCode.PERMISSION_DENIED,
            ]:
                raise DoNotRetryException(f"Code: {e.code()} | {e.details()}")  # type: ignore
            else:
                raise e
        artifacts = process_artifacts_from_answers(
            prefix=prefix,
            prompt=prompt,  # type: ignore
            answers=answers,
            write=False,
            verbose=False,
        )
        out = []

        # store images
        for out_p, artifact in artifacts:
            if artifact.type == ARTIFACT_IMAGE:  # type: ignore
                data_bin = bytes(artifact.binary)  # type: ignore
                out_p = out_p.replace("/", "--")
                fp = store_blob(out_p, data_bin)
                out.append(fp)
        return out

    return exponential_backoff(_fn, max_retries=max_retries, retry_delay=retry_delay)  # type: ignore


model_registry.register(
    model=Model(
        collection_name="stabilityai",
        id="stability-text-to-image",
        fn=stability_text_to_image,
        description="Generate a new image from a text prompt",
    )
)



from dataclasses import dataclass
from typing import Dict, Any, List, Optional

from pydantic import BaseModel


class FENode(BaseModel):
    class CFData(BaseModel):
        id: str
        type: str
        node: Dict[str, Any]
        value: Any = None

    class Position(BaseModel):
        x: float
        y: float

    id: str
    cf_id: str = ""  # this is the id of the node in the chainfury graph
    cf_data: Optional[CFData] = None  # this is the data of the node in the chainfury graph
    position: Position
    type: str
    width: int
    height: int
    selected: bool = None  # type: ignore
    position_absolute: Position = None  # type: ignore
    dragging: bool = None  # type: ignore
    data: dict = {}


class Edge(BaseModel):
    id: str
    source: str
    sourceHandle: str = ""
    target: str
    targetHandle: str = ""


class Dag(BaseModel):
    nodes: List[FENode]
    edges: List[Edge]
    sample: Dict[str, Any] = {}  # type: ignore
    main_in: str = ""
    main_out: str = ""



try:
    from pymilvus import connections, FieldSchema, CollectionSchema, DataType, Collection, utility

    MILVUS_CLIENT_INSTALLED = True
except ImportError:
    MILVUS_CLIENT_INSTALLED = False


class MilvusVectorDBComponent:
    def __init__(self, host: str = "127.0.0.1", port: str = "19530"):
        connections.connect(host=host, port=port)
        pass

    def write(self, collection_name: str):
        pass



from uuid import uuid4
from functools import lru_cache
from typing import List, Dict, Tuple, Optional, Union

try:
    from qdrant_client import models, QdrantClient

    QDRANT_CLIENT_INSTALLED = True
except ImportError:
    QDRANT_CLIENT_INSTALLED = False

from chainfury import Secret, memory_registry, logger
from chainfury.components.const import Env, ComponentMissingError


@lru_cache(maxsize=1)
def _get_qdrant_client(url: str, api_key: str):
    """Create a qdrant client and cache it

    Args:
        url (str): qdrant url
        api_key (str): qdrant api key

    Returns:
        qdrant_client.QdrantClient: qdrant client
    """
    logger.info("Creating Qdrant client")
    return QdrantClient(url=url, api_key=api_key)


def qdrant_write(
    embeddings: List[List[float]],
    collection_name: str,
    qdrant_url: Secret = Secret(""),
    qdrant_api_key: Secret = Secret(""),
    extra_payload: List[Dict[str, str]] = [],
    wait: bool = True,
    create_if_not_present: bool = True,
    distance: str = "cosine",
) -> Tuple[str, Optional[Exception]]:
    """
    Write to the Qdrant DB using the Qdrant client. In order to use this, access via the `memory_registry`:

    Example:
        >>> from chainfury import memory_registry
        >>> mem = memory_registry.get_write("qdrant")
        >>> sentence = "C.P. Cavafy is widely considered the most distinguished Greek poet of the 20th century."
        >>> out, err = mem(
                {
                    "items": [sentence],
                    "extra_payload": [
                        {"data": sentence},
                    ],
                    "collection_name": "my_test_collection",
                    "embedding_model": "openai-embedding",
                    "create_if_not_present": True,
                }
            )
        >>> if err:
                print("TRACE:", out)
            else:
                print(out)

    Args:
        embeddings (List[List[float]]): list of embeddings
        collection_name (str): collection name
        qdrant_url (Secret, optional): qdrant url or set env var `QDRANT_API_URL`.
        qdrant_api_key (Secret, optional): qdrant api key or set env var `QDRANT_API_KEY`.
        extra_payload (List[Dict[str, str]], optional): extra payload. Defaults to [].
        wait (bool, optional): wait for the response. Defaults to True.
        create_if_not_present (bool, optional): create collection if not present. Defaults to True.
        distance (str, optional): distance metric. Defaults to "cosine".

    Returns:
        Tuple[str, Optional[Exception]]: status and error
    """
    # client check
    if not QDRANT_CLIENT_INSTALLED:
        raise ComponentMissingError("Qdrant client is not installed. Please install it with `pip install qdrant-client`")
    qdrant_url = Secret(Env.QDRANT_API_URL(qdrant_url.value)).value
    qdrant_api_key = Secret(Env.QDRANT_API_KEY(qdrant_api_key.value)).value
    if not qdrant_url:
        raise Exception("Qdrant URL is not set. Please pass `qdrant_url` or  env var `QDRANT_API_URL=<your_url>`")
    if not qdrant_api_key:
        raise Exception("Qdrant API KEY is not set. Please pass `qdrant_api_key` or  env var `QDRANT_API_KEY=<your_url>`")

    # checks
    if not (len(embeddings) and len(embeddings[0]) and type(embeddings[0][0]) == float):
        raise Exception("Embeddings should be a list of lists of floats")
    if extra_payload and len(extra_payload) != len(embeddings):
        raise Exception("Length of extra_payload should be equal to embeddings")

    client: QdrantClient = _get_qdrant_client(qdrant_url, qdrant_api_key)

    # next we create points and upsert them into the DB
    points = []
    for i, embedding in enumerate(embeddings):
        payload = {}
        if extra_payload:
            payload = extra_payload[i]
        points.append(models.PointStruct(id=str(uuid4()), payload=payload, vector=embedding))
    batch = models.Batch(
        ids=[point.id for point in points],
        vectors=[point.vector for point in points],
        payloads=[point.payload for point in points],
    )

    def _insert():
        try:
            result = client.upsert(
                collection_name=collection_name,
                points=batch,
                wait=wait,
            )
        except Exception as e:
            return e.content, e  # type: ignore
        return result.status.lower(), None

    status, err = _insert()
    if err and err.status_code == 404 and create_if_not_present:  # type: ignore
        collection = client.recreate_collection(
            collection_name=collection_name,
            vectors_config=models.VectorParams(
                size=len(embeddings[0]),
                distance=getattr(models.Distance, distance.upper()),
            ),
        )
        logger.info(f"Created collection {collection}")
        status, err = _insert()
    return status, err


memory_registry.register_write(
    component_name="qdrant",
    fn=qdrant_write,
    outputs={"status": 0},
    vector_key="embeddings",
)


def qdrant_read(
    embeddings: List[List[float]],
    collection_name: str,
    cutoff_score: float = 0.0,
    limit: int = 3,
    offset: int = 0,
    qdrant_url: Secret = Secret(""),
    qdrant_api_key: Secret = Secret(""),
    qdrant_search_hnsw_ef: int = 0,
    qdrant_search_exact: bool = False,
    batch_search: bool = False,
) -> Tuple[Dict[str, List[Dict[str, Union[float, int]]]], Optional[Exception]]:
    """
    Read from the Qdrant DB using the Qdrant client. In order to use this access via the `memory_registry`:

    Example:
        >>> from chainfury import memory_registry
        >>> mem = memory_registry.get_read("qdrant")
        >>> sentence = "Who was the Cafavy?"
        >>> out, err = mem(
                {
                    "items": [sentence],
                    "collection_name": "my_test_collection",
                    "embedding_model": "openai-embedding"
                }
            )
        >>> if err:
                print("TRACE:", out)
            else:
                print(out)

    Note:
        `batch_search` is not implemented yet. There's some issues from the `qdrant_client` library.

    Args:
        embeddings (List[List[float]]): list of embeddings
        collection_name (str): collection name
        cutoff_score (float, optional): cutoff score. Defaults to 0.0.
        limit (int, optional): limit. Defaults to 3.
        offset (int, optional): offset. Defaults to 0.
        qdrant_url (Secret, optional): qdrant url or set env var `QDRANT_API_URL`.
        qdrant_api_key (Secret, optional): qdrant api key or set env var `QDRANT_API_KEY`.
        qdrant_search_hnsw_ef (int, optional): qdrant search beam size, the larger the beam size the more accurate the search,
            if not set uses default value.
        qdrant_search_exact (bool, optional): qdrant search exact. Defaults to False.
        batch_search (bool, optional): batch search. Defaults to False.

    Returns:
        Tuple[List[Dict[str, Union[float, int]]], Optional[Exception]]: list of results and error
    """
    # client check
    if not QDRANT_CLIENT_INSTALLED:
        raise ComponentMissingError("Qdrant client is not installed. Please install it with `pip install qdrant-client`")
    qdrant_url = Secret(Env.QDRANT_API_URL(qdrant_url.value)).value
    qdrant_api_key = Secret(Env.QDRANT_API_KEY(qdrant_api_key.value)).value
    if not qdrant_url:
        raise Exception("Qdrant URL is not set. Please pass `qdrant_url` or  env var `QDRANT_API_URL=<your_url>`")
    if not qdrant_api_key:
        raise Exception("Qdrant API KEY is not set. Please pass `qdrant_api_key` or  env var `QDRANT_API_KEY=<your_url>`")

    # checks
    if not (len(embeddings) and len(embeddings[0]) and type(embeddings[0][0]) == float):
        raise Exception("Embeddings should be a list of lists of floats")
    if batch_search:
        raise NotImplementedError("Batch search is not implemented yet")
    if not batch_search and len(embeddings) > 1:
        raise Exception("Batch search is not enabled, but multiple embeddings are passed")

    client: QdrantClient = _get_qdrant_client(qdrant_url, qdrant_api_key)

    search_params = models.SearchParams()
    if qdrant_search_hnsw_ef:
        search_params.hnsw_ef = qdrant_search_hnsw_ef
    if qdrant_search_exact:
        search_params.exact = qdrant_search_exact

    if batch_search:
        # this is not implemented, this fails when we try to pass a list of vectors
        search_queries = [models.SearchRequest(vector=x, limit=limit, offset=offset, params=search_params) for x in embeddings]
        out = client.search_batch(
            collection_name=collection_name,
            requests=search_queries,
        )
        res = [[_x.dict(skip_defaults=False) for _x in x] for x in out]

    out = client.search(
        collection_name=collection_name,
        query_vector=embeddings[0],
        limit=limit,
        offset=offset,
        search_params=search_params,
    )
    out = [x for x in out if x.score > cutoff_score]
    res = [_x.dict(skip_defaults=False) for _x in out]
    return {"data": res}, None


memory_registry.register_read(
    component_name="qdrant",
    fn=qdrant_read,
    outputs={"items": 0},
    vector_key="embeddings",
)



import json
import requests
from typing import Dict, Tuple, Optional, List

from chainfury import programatic_actions_registry, exponential_backoff, Secret, UnAuthException
from chainfury.components.const import Env


VALID_SEARCH_TYPES = [
    "search",
    "images",
    "places",
    "news",
]


def serper_api(
    query: str,
    serper_api_key: Secret = Secret(""),
    search_type: str = "search",
    location: str = "in",
    locale: str = "en",
    autocorrect: bool = True,
    page: int = 1,
    num_per_page: int = 10,
    retry_count: int = 3,
    retry_delay: int = 1,
) -> Tuple[Tuple[str, int], Optional[Exception]]:
    """
    Search the web with Serper.

    **NOTE**: This action requires a Serper API key. You can get one for free at https://serper.dev.

    Here's a few types of responses that the different serper APIs return (this was written on 3rd July, 2023):

    .. code-block:: python

        >>> news_api
        {
            "news": [
                {
                    "date": "2 hours ago",
                    "imageUrl": "https://encrypted-tbn0.gstatic.com/images?q=tbn:ANd9GcQ0KmSaKaJRif6f0wIBvRfnQfcQOXnoWWQ9ZM0VfTz9n6NXbCEUp76Nj-z1bw&s",
                    "link": "https://sg.news.yahoo.com/apple-makes-history-first-3-085543504.html",
                    "position": 1,
                    "snippet": "(Reuters) - Apple Inc become the first company in the world to reach a market value of $3 trillion, buoyed by hopes over its expansion in...",
                    "source": "Yahoo News Singapore",
                    "title": "Apple makes history as first $3 trillion company amid tech stock surge"
                }
            ],
            "searchParameters": {
                "autocorrect": true,
                "gl": "us",
                "hl": "en",
                "num": 10,
                "page": 1,
                "q": "apple inc",
                "type": "news"
            }
        }
        >>> search_api
        {
            "searchParameters": {
                "q": "apple inc",
                "gl": "us",
                "hl": "en",
                "num": 20,
                "autocorrect": false,
                "page": 3,
                "type": "search"
            },
            "organic": [
                {
                    "title": "Chennai, Tamil Nadu, India Weather Forecast - AccuWeather",
                    "link": "https://www.accuweather.com/en/in/chennai/206671/weather-forecast/206671",
                    "snippet": "Chennai, Tamil Nadu ; Current Weather. 4:17 PM. 82°F · RealFeel® 91° ; TODAY'S WEATHER FORECAST. 7/3. 83°Hi. RealFeel® 95° ; TONIGHT'S WEATHER FORECAST. 7/3. 78°Lo.",
                    "position": 1
                }
            ],
            "answerBox": {
                "title": "Chennai, Tamil Nadu, India / Weather",
                "answer": "87°F"
            },
            "knowledgeGraph": {
                "title": "Jeff Bezos",
                "type": "Executive Chairman of Amazon",
                "imageUrl": "https://encrypted-tbn0.gstatic.com/images?q=tbn:ANd9GcTxFsGLOkNFB1vs2kvXM1_WNJZOwWHtdGaaLAnsHA&s=0",
                "description": "Jeffrey Preston Bezos is an American entrepreneur, media proprietor, and investor. He is the founder, executive chairman, and former president and CEO of Amazon, the world's largest e-commerce and cloud computing company.",
                "descriptionSource": "Wikipedia",
                "descriptionLink": "https://en.wikipedia.org/wiki/Jeff_Bezos",
                "attributes": {
                    "Born": "12 January 1964 (age 59 years), Albuquerque, New Mexico, United States",
                    "Net worth": "15,730 crores USD (2023)",
                    "Spouse": "MacKenzie Scott (m. 1993–2019)",
                    "Children": "Preston Bezos",
                    "Parents": "Ted Jorgensen, Miguel Bezos and Jacklyn Bezos",
                    "Height": "1.71 m",
                    "Nationality": "American"
                }
            }
            "peopleAlsoAsk": [
                {
                    "question": "What is the climate now in Chennai?",
                    "snippet": "The weather today in Chennai will be very hot with temperatures reaching 96°F.",
                    "title": "The local weather in Chennai today - Weather25.com",
                    "link": "https://www.weather25.com/asia/india/tamil-nadu/chennai?page=today"
                }
            ],
            "relatedSearches": [
                {
                    "query": "What is the weather in chennai this week"
                },
                {
                    "query": "Weather forecast Chennai cyclone"
                }
            ]
        }
        >>> image_api
        {
            "images": [
                {
                    "title": "Chennai weather forecast: Chennai, other parts of Tamil Nadu ...",
                    "imageUrl": "https://static.toiimg.com/photo/msid-71699845/71699845.jpg",
                    "imageWidth": 1200,
                    "imageHeight": 900,
                    "thumbnailUrl": "https://encrypted-tbn0.gstatic.com/images?q=tbn:ANd9GcTscW9XwTVb2S1aHCgaDHgvdTKHP6Vq8rAk4F9dm6YP8udSa6rZ&amp;s",
                    "thumbnailWidth": 259,
                    "thumbnailHeight": 194,
                    "source": "Times of India",
                    "domain": "timesofindia.indiatimes.com",
                    "link": "https://timesofindia.indiatimes.com/city/chennai/chennai-other-parts-of-tamil-nadu-likely-to-get-heavy-to-very-heavy-rain-for-next-three-days/articleshow/71699802.cms",
                    "position": 1
                }
            ],
            "searchParameters": {
                "q": "what is the weather in chennai?",
                "gl": "us",
                "hl": "en",
                "num": 20,
                "autocorrect": false,
                "page": 1,
                "type": "images"
            }
        }


    Args:
        query (str): The search query.
        serper_api_key (Secret): The Serper API key. Defaults to the value of SERPER_API_KEY environment variable.
        search_type (str): The type of search to perform. Must be one of "search", "images", "places", or "news".
        resolver (Dict[str, Union[str, List[str], Tuple[str, ...]]]): The keys to return in the output JSON.
        location (str): The location to search from. Defaults to "in".
        locale (str): The locale to search from. Defaults to "en".
        autocorrect (bool): Whether to autocorrect the query. Defaults to True.
        page (int): The page number to return. Defaults to 1.
        num_per_page (int): The number of results to return per page. Defaults to 10.
        retry_count (int): The number of times to retry the request. Defaults to 3.
        retry_delay (int): The number of seconds to wait between retries. Defaults to 1.

    Returns:
        dict: The search results.
    """
    if search_type not in VALID_SEARCH_TYPES:
        raise ValueError(f"search_type must be one of {VALID_SEARCH_TYPES}")
    if not serper_api_key:
        serper_api_key = Secret(Env.SERPER_API_KEY("")).value
    if not serper_api_key:
        raise Exception("Serper API key not found. Please set SERPER_API_KEY environment variable or pass it as an argument.")

    def _fn():
        r = requests.post(
            f"https://google.serper.dev/search",
            headers={
                "X-API-KEY": serper_api_key,
                "Content-Type": "application/json",
            },
            json={
                "q": query,
                "gl": location,
                "hl": locale,
                "num": num_per_page,
                "autocorrect": autocorrect,
                "page": page,
                "type": search_type,
            },
        )
        if r.status_code == 401:
            raise UnAuthException(r.text)
        elif r.status_code == 403:
            raise Exception("Serper API key is invalid.")
        elif r.status_code != 200:
            raise Exception(f"Serper API returned status code {r.status_code}: {r.text}")

        return r.text, r.status_code

    try:
        text, status_code = exponential_backoff(foo=_fn, max_retries=retry_count, retry_delay=retry_delay)
        return (text, status_code), None  # type: ignore
    except Exception as e:
        return (None, None), e  # type: ignore


programatic_actions_registry.register(
    fn=serper_api,
    node_id="serper-api",
    description="Search the web with Serper",
    outputs={
        "text": (0,),
        "status_code": (1,),
    },
)



import requests
from pydantic import BaseModel
from typing import Any, List, Union, Dict, Optional

from chainfury import Secret, model_registry, exponential_backoff, Model, UnAuthException
from chainfury.components.const import Env


def openai_completion(
    model: str,
    prompt: Union[str, List[Union[str, List[str]]]],
    openai_api_key: Secret = Secret(""),
    max_tokens: int = 16,
    temperature: float = 1,
    top_p: float = 1,
    n: int = 1,
    logprobs: int = 0,
    echo: bool = False,
    stop: Union[str, List[str]] = "",
    presence_penalty: float = 0,
    frequency_penalty: float = 0,
    best_of: int = 1,
    logit_bias: dict = {},
    user: str = "",
    *,
    retry_count: int = 3,
    retry_delay: int = 1,
) -> Dict[str, Any]:
    """
    Generate text completion using OpenAI's GPT API.

    Args:
        model (str): ID of the model to use.
        prompt (Union[str, List[Union[str, List[str]]]]): The prompt(s) to generate completions for.
            Encoded as a string, array of strings, array of tokens, or array of token arrays.
        openai_api_key (Secret): The OpenAI API key. Defaults to "" or the OPENAI_TOKEN environment variable.
        max_tokens (Optional[int]): The maximum number of tokens to generate in the completion. Defaults to 16.
        temperature (Optional[float]): What sampling temperature to use, between 0 and 2. Defaults to 1.
        top_p (Optional[float]): An alternative to sampling with temperature. Defaults to 1.
        n (Optional[int]): How many completions to generate for each prompt. Defaults to 1.
        logprobs (Optional[int]): Include the log probabilities on the logprobs most likely tokens.
            The maximum value for logprobs is 5. Defaults to None.
        echo (Optional[bool]): Echo back the prompt in addition to the completion. Defaults to False.
        stop (Optional[Union[str, List[str]]]): Up to 4 sequences where the API will stop generating further tokens.
            The returned text will not contain the stop sequence. Defaults to None.
        presence_penalty (Optional[float]): Number between -2.0 and 2.0. Positive values penalize new tokens based on
            whether they appear in the text so far, increasing the model's likelihood to talk about new topics.
            Defaults to 0.
        frequency_penalty (Optional[float]): Number between -2.0 and 2.0. Positive values penalize new tokens based on
            their existing frequency in the text so far, decreasing the model's likelihood to repeat the same line
            verbatim. Defaults to 0.
        best_of (Optional[int]): Generates best_of completions server-side and returns the "best".
            Results cannot be streamed. Defaults to 1.
        logit_bias (Optional[dict]): Modify the likelihood of specified tokens appearing in the completion.
            Accepts a json object that maps tokens (specified by their token ID in the GPT tokenizer) to an associated
            bias value from -100 to 100. Defaults to None.
        user (Optional[str]): A unique identifier representing your end-user, which can help OpenAI to monitor and detect
            abuse. Defaults to None.
        retry_count: Optional. Number of times to retry the API call. Defaults to 3.
        retry_delay: Optional. Number of seconds to wait before retrying the API call. Defaults to 1.

    Returns:
        Any: The completion(s) generated by the API.

    """
    if not openai_api_key:
        openai_api_key = Secret(Env.OPENAI_TOKEN("")).value
    if not openai_api_key:
        raise Exception("OpenAI API key not found. Please set OPENAI_TOKEN environment variable or pass through function")

    def _fn():
        r = requests.post(
            "https://api.openai.com/v1/completions",
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {openai_api_key}",
            },
            json={
                "model": model,
                "prompt": prompt,
                "max_tokens": max_tokens,
                "temperature": temperature,
                "top_p": top_p,
                "n": n,
                "logprobs": logprobs,
                "echo": echo,
                "stop": stop,
                "presence_penalty": presence_penalty,
                "frequency_penalty": frequency_penalty,
                "best_of": best_of,
                "logit_bias": logit_bias,
                "user": user,
            },
        )
        if r.status_code == 401:
            raise UnAuthException(r.text)
        if r.status_code != 200:
            raise Exception(f"OpenAI API returned status code {r.status_code}: {r.text}")
        return r.json()

    return exponential_backoff(_fn, max_retries=retry_count, retry_delay=retry_delay)


model_registry.register(
    model=Model(
        collection_name="openai",
        id="openai-completion",
        fn=openai_completion,
        description="Given a prompt, the model will return one or more predicted completions, and can also return the probabilities of alternative tokens at each position.",
        usage=["usage", "total_tokens"],
    ),
)


class OpenAIChat(BaseModel):
    """This model is used to chat with the OpenAI API"""

    class Message(BaseModel):
        role: str
        content: str
        name: Optional[str] = None
        function_call: Optional[Union[str, Dict[str, str]]] = None

    class Function(BaseModel):
        name: str
        description: Optional[str] = None
        parameters: Dict[str, str]
        function_call: Optional[str] = None

    model: str
    messages: List[Message]
    functions: Optional[List[Function]] = None
    temperature: Optional[float] = 1
    top_p: Optional[float] = 1
    n: Optional[int] = 1
    stream: Optional[bool] = False
    stop: Optional[Union[str, List[str]]] = None
    max_tokens: Optional[int] = 2 << 32 - 1
    presence_penalty: Optional[float] = 0
    frequency_penalty: Optional[float] = 0
    logit_bias: Optional[Dict[str, int]] = None
    user: Optional[str] = None


def openai_chat(
    model: str,
    messages: List[Dict[str, str]],
    openai_api_key: Secret = Secret(""),
    temperature: float = 1.0,
    top_p: float = 1.0,
    n: int = 1,
    stop: Union[str, List[str]] = "",
    max_tokens: int = 1024,
    presence_penalty: float = 0.0,
    frequency_penalty: float = 0.0,
    logit_bias: dict = {},
    user: str = "",
    *,
    retry_count: int = 3,
    retry_delay: int = 1,
) -> Any:
    """
    Returns a JSON object containing the OpenAI's API chat response.

    Args:
        model: ID of the model to use. See the model endpoint compatibility table for details on which models work with the Chat API.
        messages: A list of messages describing the conversation so far, each item contains the folowing keys
            role: The role of the author of this message. One of system, user, or assistant.
            content: The contents of the message.
            name: Optional. The name of the author of this message. May contain a-z, A-Z, 0-9, and underscores, with a maximum length of 64 characters.
        openai_api_key (Secret): The OpenAI API key. Defaults to "" or the OPENAI_TOKEN environment variable.
        temperature: Optional. What sampling temperature to use, between 0 and 2. Higher values like 0.8 will make the output more random, while lower values like 0.2 will make it more focused and deterministic. We generally recommend altering this or top_p but not both. Defaults to 1.
        top_p: Optional. An alternative to sampling with temperature, called nucleus sampling, where the model considers the results of the tokens with top_p probability mass. So 0.1 means only the tokens comprising the top 10% probability mass are considered. We generally recommend altering this or temperature but not both. Defaults to 1.
        n: Optional. How many chat completion choices to generate for each input message. Defaults to 1.
        stop: Optional. Up to 4 sequences where the API will stop generating further tokens.
        max_tokens: Optional. The maximum number of tokens to generate in the chat completion. The total length of input tokens and generated tokens is limited by the model's context length. Defaults to infinity.
        presence_penalty: Optional. Number between -2.0 and 2.0. Positive values penalize new tokens based on whether they appear in the text so far, increasing the model's likelihood to talk about new topics. See more information about frequency and presence penalties. Defaults to 0.
        frequency_penalty: Optional. Number between -2.0 and 2.0. Positive values penalize new tokens based on their existing frequency in the text so far, decreasing the model's likelihood to repeat the same line verbatim. See more information about frequency and presence penalties. Defaults to 0.
        logit_bias: Optional. Modify the likelihood of specified tokens appearing in the completion. Accepts a json object that maps tokens (specified by their token ID in the tokenizer) to an associated bias value from -100 to 100. Mathematically, the bias is added to the logits generated by the model prior to sampling. The exact effect will vary per model, but values between -1 and 1 should decrease or increase likelihood of selection; values like -100 or 100 should result in a ban or exclusive selection of the relevant
        user: Optional. A unique identifier representing your end-user, which can help OpenAI to monitor and detect abuse. Defaults to None.
        retry_count: Optional. Number of times to retry the API call. Defaults to 3.
        retry_delay: Optional. Number of seconds to wait before retrying the API call. Defaults to 1.

    Returns:
        Any: The completion(s) generated by the API.
    """
    if not openai_api_key:
        openai_api_key = Secret(Env.OPENAI_TOKEN("")).value
    if not openai_api_key:
        raise Exception("OpenAI API key not found. Please set OPENAI_TOKEN environment variable or pass through function")

    if not len(messages):
        raise Exception("Messages cannot be empty")
    if isinstance(messages[0], OpenAIChat.Message):
        messages = [x.dict(skip_defaults=True) for x in messages]

    def _fn():
        r = requests.post(
            "https://api.openai.com/v1/chat/completions",
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {openai_api_key}",
            },
            json={
                "model": model,
                "messages": messages,
                "max_tokens": max_tokens,
                "temperature": temperature,
                "top_p": top_p,
                "n": n,
                "stop": stop,
                "presence_penalty": presence_penalty,
                "frequency_penalty": frequency_penalty,
                "logit_bias": logit_bias,
                "user": user,
            },
        )
        if r.status_code == 401:
            raise UnAuthException(r.text)
        if r.status_code != 200:
            raise Exception(f"OpenAI API returned status code {r.status_code}: {r.text}")
        return r.json()

    return exponential_backoff(_fn, max_retries=retry_count, retry_delay=retry_delay)


model_registry.register(
    model=Model(
        collection_name="openai",
        id="openai-chat",
        fn=openai_chat,
        description="Given a list of messages describing a conversation, the model will return a response.",
        usage=["usage", "total_tokens"],
    )
)


def openai_embedding(
    model: str,
    input_strings: Union[str, List[str]],
    openai_api_key: Secret = Secret(""),
    user: str = "",
    *,
    retry_count: int = 3,
    retry_delay: int = 1,
) -> Any:
    """
    Returns a JSON object containing the OpenAI's embeddings API response

    Args:
        model: ID of the model to use. See the model endpoint compatibility table for details on which models work with the Embeddings API.
        input_strings: A list of strings to embed, encoded as a string, array of strings, array of tokens, or array of token arrays.
        openai_api_key (Secret): The OpenAI API key. Defaults to "" or the OPENAI_TOKEN environment variable.
        user: Optional. A unique identifier representing your end-user, which can help OpenAI to monitor and detect abuse. Defaults to None.
        retry_count: Optional. Number of times to retry the API call. Defaults to 3.
        retry_delay: Optional. Number of seconds to wait before retrying the API call. Defaults to 1.

    Returns:
        Any: The completion(s) generated by the API.
    """

    if not openai_api_key:
        openai_api_key = Secret(Env.OPENAI_TOKEN("")).value
    if not openai_api_key:
        raise Exception("OpenAI API key not found. Please set OPENAI_TOKEN environment variable or pass through function")

    def _fn():
        r = requests.post(
            "https://api.openai.com/v1/embeddings",
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {openai_api_key}",
            },
            json={
                "model": model,
                "input": input_strings,
                "user": user,
            },
        )
        if r.status_code == 401:
            raise UnAuthException(r.text)
        if r.status_code != 200:
            raise Exception(f"OpenAI API returned status code {r.status_code}: {r.text}")
        return r.json()

    return exponential_backoff(_fn, max_retries=retry_count, retry_delay=retry_delay)


model_registry.register(
    model=Model(
        collection_name="openai",
        id="openai-embedding",
        fn=openai_embedding,
        description="Given a list of messages create embeddings for each message.",
        usage=["usage", "total_tokens"],
    )
)



import os
import time
from typing import List

try:
    import pinecone

    PINECONE_CLIENT_INSTALLED = True
except ImportError:
    PINECONE_CLIENT_INSTALLED = False

from chainfury import model_registry, Secret
from chainfury.components.const import Env, ComponentMissingError


def write_pinecone(
    items: List[str],
    index_name: str,
    pinecone_api_key: Secret = Secret(""),
    piencone_env: Secret = Secret(""),
    embedding_model: str = "openai-embedding",
    model_data: dict = {},
    metric: str = "cosine",
    create_if_not_present: bool = False,
):
    pinecone.init(
        api_key=Env.PINECONE_API_KEY(pinecone_api_key.value),
        environment=Env.PINECONE_ENV(piencone_env.value),
    )

    # create an embeddings first for one item so in case the index does not exist we can create it by passing the
    # embedding dimension
    model = model_registry.get(embedding_model)
    out = model(model_data)

    # only create index if it doesn't exist
    if index_name not in pinecone.list_indexes():
        if create_if_not_present:
            pinecone.create_index(
                name=index_name,
                dimension=len(dataset.documents.iloc[0]["values"]),
                metric=metric,
                shards=1,
                index_file_size=1024,
            )
            time.sleep(1)  # wait a moment for the index to be fully initialized
        else:
            raise Exception(
                f"Index {index_name} does not exist. Please create it first or set create_if_not_present=True",
            )


def read_pinecone(
    query: str,
    index_name: str,
    pinecone_api_key: Secret = Secret(""),
    piencone_env: Secret = Secret(""),
    embedding_model: str = "openai-embedding",
    metric: str = "cosine",
    create_if_not_present: bool = False,
    model_data: dict = {},
):
    pinecone.init(
        api_key=Env.PINECONE_API_KEY(pinecone_api_key.value),
        environment=Env.PINECONE_ENV(piencone_env.value),
    )

    # create an embeddings first for one item so in case the index does not exist we can create it by passing the
    # embedding dimension
    model = model_registry.get(embedding_model)
    out = model(model_data)

    # only create index if it doesn't exist
    if index_name not in pinecone.list_indexes():
        if create_if_not_present:
            pinecone.create_index(
                name=index_name,
                dimension=len(dataset.documents.iloc[0]["values"]),
                metric=metric,
                shards=1,
                index_file_size=1024,
            )
            time.sleep(1)  # wait a moment for the index to be fully initialized
        else:
            raise Exception(
                f"Index {index_name} does not exist. Please create it first or set create_if_not_present=True",
            )



from chainfury.base import Var, Node, Secret, Chain, Model, Edge
from chainfury.agent import model_registry, programatic_actions_registry, ai_actions_registry, memory_registry, AIAction, Memory
from chainfury.utils import exponential_backoff, UnAuthException, DoNotRetryException, logger
from chainfury.client import get_client
from chainfury import components



import os


class Env:
    """
    Single namespace for all environment variables. This performs a left merge so it will prefer the value passed in
    over the environment variable.

    * CF_TOKEN: ChainFury API token
    * CF_URL: ChainFury API URL
    * NBX_DEPLOY_URL: NimbleBox Deploy URL
    * NBX_DEPLOY_KEY: NimbleBox Deploy API key

    * OPENAI_TOKEN: OpenAI API token, see platform.openai.com
    * SERPER_API_KEY: Serper API key, see serper.dev/
    * STABILITY_KEY: Stability API key, see dreamstudio.ai
    * PINECONE_API_KEY & PINECONE_ENVIRONMENT: Pinecone secrets, see https://pinecone.io/
    * QDRANT_API_URL & QDRANT_API_KEY: Qdrant secrets, see https://qdrant.tech/
    """

    # when you want to use chainfury as a client you need to set the following vars
    CF_TOKEN = lambda x: x or os.getenv("CF_TOKEN", "")
    CF_URL = lambda x: x or os.getenv("CF_URL", "")

    # when using NimbleBox Deploy
    NBX_DEPLOY_URL = lambda x: x or os.getenv("NBX_DEPLOY_URL", "")
    NBX_DEPLOY_KEY = lambda x: x or os.getenv("NBX_DEPLOY_KEY", "")

    ## different keys for different 3rd party APIs
    OPENAI_TOKEN = lambda x: x or os.getenv("OPENAI_TOKEN", "")
    SERPER_API_KEY = lambda x: x or os.getenv("SERPER_API_KEY", "")
    STABILITY_KEY = lambda x: x or os.getenv("STABILITY_KEY", "")

    # pinecone
    PINECONE_API_KEY = lambda x: x or os.getenv("PINECONE_API_KEY", "")
    PINECONE_ENV = lambda x: x or os.getenv("PINECONE_ENVIRONMENT", "")

    # qdrant
    QDRANT_API_URL = lambda x: x or os.getenv("QDRANT_API_URL", "")
    QDRANT_API_KEY = lambda x: x or os.getenv("QDRANT_API_KEY", "")


class ComponentMissingError(Exception):
    pass



import os
import time
import logging
from typing import Any, Dict


class CFEnv:
    """
    Single namespace for all environment variables.

    * CF_FOLDER: database connection string
    """

    CF_FOLDER = lambda: os.path.expanduser(os.getenv("CF_FOLDER", "~/cf"))
    CF_BLOB_STORAGE = lambda: os.path.join(CFEnv.CF_FOLDER(), "blob")


os.makedirs(CFEnv.CF_FOLDER(), exist_ok=True)
os.makedirs(CFEnv.CF_BLOB_STORAGE(), exist_ok=True)


def store_blob(key: str, value: bytes) -> str:
    """A function that stores the information in a file."""
    fp = os.path.join(CFEnv.CF_BLOB_STORAGE(), key)
    with open(fp, "wb") as f:
        f.write(value)
    return fp


def terminal_top_with_text(msg: str = "") -> str:
    """Prints full wodth text message on the terminal

    Args:
        msg (str, optional): The message to print. Defaults to "".

    Returns:
        str: The message to print
    """
    width = os.get_terminal_size().columns
    if len(msg) > width - 5:
        x = "=" * width
        x += "\n" + msg
        x += "\n" + "=" * width // 2  # type: ignore
    else:
        x = "=" * (width - len(msg) - 1) + " " + msg
    return x


def get_logger() -> logging.Logger:
    """Returns a logger object"""
    logger = logging.getLogger("fury")
    lvl = os.getenv("FURY_LOG_LEVEL", "info").upper()
    logger.setLevel(getattr(logging, lvl))
    log_handler = logging.StreamHandler()
    log_handler.setFormatter(
        logging.Formatter("[%(asctime)s] [%(levelname)s] [%(filename)s:%(lineno)d] %(message)s", datefmt="%Y-%m-%dT%H:%M:%S%z")
    )
    logger.addHandler(log_handler)
    return logger


logger = get_logger()
"""
This is the logger object that should be used across the entire package as well as by the user what wants to leverage
existing logging infrastructure.
"""


class UnAuthException(Exception):
    """Raised when the API returns a 401"""


class DoNotRetryException(Exception):
    """Raised when code tells not to retry"""


def exponential_backoff(foo, *args, max_retries=2, retry_delay=1, **kwargs) -> Dict[str, Any]:
    """Exponential backoff function

    Args:
        foo (function): The function to call
        max_retries (int, optional): maximum number of retries. Defaults to 2.
        retry_delay (int, optional): Initial delay in seconds. Defaults to 1.

    Raises:
        e: Max retries reached. Exiting...
        Exception: This should never happen

    Returns:
        Dict[str, Any]: The completion(s) generated by the API.
    """

    for attempt in range(max_retries):
        try:
            out = foo(*args, **kwargs)  # Call the function that may crash
            return out  # If successful, break out of the loop and return
        except DoNotRetryException as e:
            raise e
        except UnAuthException as e:
            raise e
        except Exception as e:
            logger.warning(f"Function crashed: {e}")
            if attempt == max_retries - 1:
                logger.error("Max retries reached. Exiting...")
                raise e
            else:
                delay = retry_delay * (2**attempt)  # Calculate the backoff delay
                logger.info(f"Retrying in {delay} seconds...")
                time.sleep(delay)  # Wait for the calculated delay
    raise Exception("This should never happen")


def folder(x: str) -> str:
    """get the folder of this file path"""
    return os.path.split(os.path.abspath(x))[0]


def joinp(x: str, *args) -> str:
    """convienience function for os.path.join"""
    return os.path.join(x, *args)



import random
import requests
from typing import Any, List, Optional

from chainfury import Secret, model_registry, exponential_backoff, Model, UnAuthException
from chainfury.components.const import Env


def nbx_chat_api(
    inputs: str,
    nbx_deploy_url: str = "",
    nbx_header_token: Secret = Secret(""),
    best_of: int = 1,
    decoder_input_details: bool = True,
    details: bool = True,
    do_sample: bool = True,
    max_new_tokens: int = 20,
    repetition_penalty: float = 1.03,
    return_full_text: bool = False,
    seed: int = None,  # type: ignore # see components README.md
    stop: List[str] = [],
    temperature: float = 0.5,
    top_k: int = 10,
    top_p: float = 0.95,
    truncate: int = None,  # type: ignore # see components README.md
    typical_p: float = 0.95,
    watermark: bool = True,
    *,
    retry_count: int = 3,
    retry_delay: int = 1,
) -> Any:
    """
    Returns a JSON object containing the OpenAI's API chat response.

    Args:
        inputs (str): The prompt to send to the API.
        nbx_deploy_url (str): The NBX deploy URL. Defaults to the value of NBX_DEPLOY_URL environment variable.
        nbx_header_token (Secret): The NBX header token. Defaults to the value of NBX_DEPLOY_KEY environment variable.
        best_of (int): The number of outputs to generate and return. Defaults to 1.
        decoder_input_details (bool): Whether to return the decoder input details. Defaults to True.
        details (bool): Whether to return the details. Defaults to True.
        do_sample (bool): Whether to use sampling. Defaults to True.
        max_new_tokens (int): The maximum number of tokens to generate. Defaults to 20.
        repetition_penalty (float): The repetition penalty. Defaults to 1.03.
        return_full_text (bool): Whether to return the full text. Defaults to False.
        seed (int): The seed to use for random number generation. Defaults to a random integer between 0 and 2^32 - 1.
        stop (List[str]): The stop tokens. Defaults to an empty list.
        temperature (float): The temperature. Defaults to 0.5.
        top_k (int): The top k. Defaults to 10.
        top_p (float): The top p. Defaults to 0.95.
        truncate (int): The truncate. Defaults to None.
        typical_p (float): The typical p. Defaults to 0.95.
        watermark (bool): Whether to include the watermark. Defaults to True.
        retry_count (int): The number of times to retry the API call. Defaults to 3.
        retry_delay (int): The number of seconds to wait between retries. Defaults to 1.

    Returns:
        Any: The JSON object containing the OpenAI's API chat response.
    """
    if not nbx_deploy_url:
        nbx_deploy_url = Env.NBX_DEPLOY_URL("")
    if not nbx_deploy_url:
        raise Exception("NBX_DEPLOY_URL not set, please set it in your environment or pass it as an argument")

    if not nbx_header_token:
        nbx_header_token = Secret(Env.NBX_DEPLOY_KEY("")).value
    if not nbx_header_token:
        raise Exception("NBX_DEPLOY_KEY not set, please set it in your environment or pass it as an argument")

    seed = seed or random.randint(0, 2**32 - 1)

    def _fn():
        r = requests.post(
            nbx_deploy_url + "/generate",
            headers={"NBX-KEY": nbx_header_token},
            json={
                "inputs": inputs,
                "parameters": {
                    "best_of": best_of,
                    "decoder_input_details": decoder_input_details,
                    "details": details,
                    "do_sample": do_sample,
                    "max_new_tokens": max_new_tokens,
                    "repetition_penalty": repetition_penalty,
                    "return_full_text": return_full_text,
                    "seed": seed,
                    "stop": stop,
                    "temperature": temperature,
                    "top_k": top_k,
                    "top_p": top_p,
                    "truncate": truncate,
                    "typical_p": typical_p,
                    "watermark": watermark,
                },
            },
        )
        if r.status_code == 401:
            raise UnAuthException(r.text)
        if r.status_code != 200:
            raise Exception(f"OpenAI API returned status code {r.status_code}: {r.text}")
        return r.json()

    return exponential_backoff(_fn, max_retries=retry_count, retry_delay=retry_delay)


model_registry.register(
    model=Model(
        collection_name="nbx",
        id="nbx-deploy",
        fn=nbx_chat_api,
        description="Call NimbleBox LLMOps deploy API",
    ),
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['chainfury',
 'chainfury.components',
 'chainfury.components.ai_actions',
 'chainfury.components.functional',
 'chainfury.components.milvus',
 'chainfury.components.nbx',
 'chainfury.components.openai',
 'chainfury.components.pinecone_db',
 'chainfury.components.qdrant',
 'chainfury.components.redis',
 'chainfury.components.serper',
 'chainfury.components.stability']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2==3.1.2',
 'fire==0.5.0',
 'jinja2schema==0.1.4',
 'pydantic>=1.10.7,<2.0.0',
 'requests>=2.28.2,<3.0.0']

extras_require = \
{'all': ['stability-sdk==0.8.3', 'qdrant-client==1.3.1'],
 'qdrant': ['qdrant-client==1.3.1'],
 'stability': ['stability-sdk==0.8.3']}

entry_points = \
{'console_scripts': ['cf = chainfury.cli:main',
                     'chainfury = chainfury.cli:main']}

setup_kwargs = {
    'name': 'chainfury',
    'version': '1.4.0',
    'description': 'ChainFury is a powerful tool that simplifies the creation and management of chains of prompts, making it easier to build complex chat applications using LLMs.',
    'long_description': '# 🦋 NimbleBox ChainFury\n\n[![linkcheck](https://img.shields.io/badge/Workflow-Passing-darkgreen)](https://github.com/NimbleBoxAI/ChainFury/actions)\n[![Downloads](https://static.pepy.tech/badge/chainfury)](https://pepy.tech/project/chainfury)\n[![linkcheck](https://img.shields.io/badge/Site-🦋ChainFury-lightblue)](https://chainfury.nbox.ai)\n[![License: Apache](https://img.shields.io/badge/License-Apache%20v2.0-red)](https://github.com/NimbleBoxAI/ChainFury/blob/main/LICENSE) \n[![Twitter](https://img.shields.io/twitter/url/https/twitter.com/NimbleBoxAI.svg?style=social&label=Follow%20%40NimbleBoxAI)](https://twitter.com/NimbleBoxAI)\n[![](https://dcbadge.vercel.app/api/server/KhF38hrAJ2?compact=true&style=flat)](https://discord.com/invite/KhF38hrAJ2)\n\n🦋 Build complex chat apps using LLMs in 4 clicks ⚡️ [Try it out here](https://chainfury.nbox.ai/). Used in production by [chat.nbox.ai](https://chat.nbox.ai).\n\n# Read the [Docs](https://nimbleboxai.github.io/ChainFury/index.html)\n\nThe documentation page contains all the information on using `chainfury` and `chainfury_server`.\n\n# Installation\n\nThere are two separate packages built into this repository, first is `chainfury` which contains the fury-engine for running\nthe DAGs and `chainfury_server` which contains the self hosted server for the GUI.\n\n``` bash\npip install chainfury\npip install chainfury_server\n\n# to launch the server\npython3 -m chainfury_server\n```\n\n### Run Docker\n\nEasiest way to run the server is to use docker. You can use the following command to run ChainFury:\n\n```bash\ndocker build . -f Dockerfile -t chainfury:latest\ndocker run -p 8000:8000 chainfury:latest\n```\n\nTo pass any env variables you can use the command:\n\n```bash\ndocker run --env ENV_KEY=ENV_VALUE -p 8000:8000 chainfury:latest\n```\n\nCheckout all the:\n- `component` environment variables [here](https://nimbleboxai.github.io/ChainFury/source/chainfury.components.const.html#chainfury.components.const.Env)\n- `chainfury` specific variables [here](https://nimbleboxai.github.io/ChainFury/source/chainfury.utils.html#chainfury.utils.CFEnv)\n- `chainfury_server` specific variables [here](https://nimbleboxai.github.io/ChainFury/cf_server/chainfury_server.commons.config.html#chainfury_server.commons.config.Env)\n\n### From Source\n\nHere\'s a breakdown of folder:\n\n- `chainfury/` contains the chainfury engine\n- `server/` contains the chainfury server\n- `client/` contains the frontend code for the GUI\n- `api_docs/` contains the documentation\n\nTo build the entire system from scratch follow these steps:\n\n```bash\ngit clone https://github.com/NimbleBoxAI/ChainFury\ncd ChainFury\npython3 -m venv venv\nsource venv/bin/activate\n```\n\nYou will need to have `yarn` installed to build the frontend and move it to the correct location on the server\n\n```bash\nsh stories/build_and_copy.sh\n```\n\nOnce the static files are copied we can now proceed to install dependecies:\n\n```bash\npip install setuptools\npip install -e .          # editable install the chainfury\ncd server\npip install -e .          # editable install the chainfury_server\n```\n\nTo start you can now do:\n\n```bash\ncd chainfury_server\npython3 server.py\n```\n\nYou can now visit [localhost:8000](http://localhost:8000/ui/) to see the GUI.\n\n# Contibutions\n\nChainFury is an open-source project used in production. We are open to contributions to the project in the form of features,\ninfrastructure or documentation.\n\n- Our [issues](https://github.com/NimbleBoxAI/ChainFury/issues) page is kept up to date with bugs, improvements, and feature requests.\n\n- If you\'re looking for help with your code, hop onto [GitHub Discussions board](https://github.com/NimbleBoxAI/ChainFury/discussions) or\n[Discord](https://discord.com/invite/KhF38hrAJ2), so that more people can benefit from it.\n\n- **Describing your issue:** Try to provide as many details as possible. What exactly goes wrong? How is it failing?\nIs there an error? "XY doesn\'t work" usually isn\'t that helpful for tracking down problems. Always remember to include\nthe code you ran and if possible, extract only the relevant parts and don\'t just dump your entire script. This will make\nit easier for us to reproduce the error.\n\n- **Sharing long blocks of code or logs:** If you need to include long code, logs or tracebacks, you can wrap them in\n`<details>` and `</details>`. This [collapses the content](https://developer.mozilla.org/en-US/docs/Web/HTML/Element/details)\nso it only becomes visible on click, making the issue easier to read and follow.\n\n',
    'author': 'NimbleBox Engineering',
    'author_email': 'engineering@nimblebox.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/NimbleBoxAI/ChainFury',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import os
import importlib

from chainfury.utils import logger
from chainfury.utils import folder, joinp


all_items = set()
for p in os.listdir(folder(__file__)):
    if p.startswith("_") or p in ["functional", "ai_actions"]:
        continue
    if os.path.isdir(joinp(folder(__file__), p)):
        mod = importlib.import_module(f"chainfury.components.{p}")
        logger.debug(f"Adding file: {mod.__file__}")
        all_items.add(p)
all_items = list(all_items)

# do this after the above components are loaded
from chainfury.components.functional import *
from chainfury.components.ai_actions import *



"""
Functional components are the programatic components that are available to the fury
system. These are mostly for demo examples, we expect the user to register their
unique components into programatic_action_registry.
"""

import re
import json
import requests
from typing import Any, List, Dict, Tuple, Optional, Union

from chainfury import programatic_actions_registry, exponential_backoff
from chainfury.base import get_value_by_keys

# Call API: very basic always helpful

_VALID_HTTP_METHODS = ["GET", "POST", "PUT", "DELETE", "PATCH"]


def call_api_requests(
    method: str,
    url: str,
    params: Dict[str, str] = {},
    data: Dict[str, str] = {},
    json: Dict[str, str] = {},
    headers: Dict[str, str] = {},
    cookies: Dict[str, str] = {},
    auth: Dict[str, str] = {},
    timeout: float = 0,
    *,
    max_retries: int = 3,
    retry_delay: int = 1,
) -> Tuple[Tuple[str, int], Optional[Exception]]:
    """Call an API using the python requests library. You can use this for any API that you want to call.

    Args:
        method (str): The HTTP method to use.
        url (str): The URL to call.
        params (Dict[str, str], optional): The query parameters. Defaults to {}.
        data (Dict[str, str], optional): The data to send. Defaults to {}.
        json (Dict[str, str], optional): The JSON to send. Defaults to {}.
        headers (Dict[str, str], optional): The headers to send. Defaults to {}.
        cookies (Dict[str, str], optional): The cookies to send. Defaults to {}.
        auth (Dict[str, str], optional): The auth to send. Defaults to {}.
        timeout (float, optional): The timeout in seconds. Defaults to 0.
        max_retries (int, optional): The number of times to retry the request. Defaults to 3.
        retry_delay (int, optional): The number of seconds to wait between retries. Defaults to 1.

    Returns:
        Tuple[Tuple[str, int], Optional[Exception]]: The response text and status code, and the exception if there was one.
    """
    method = method.upper()
    if method not in _VALID_HTTP_METHODS:
        raise ValueError(f"method must be one of {_VALID_HTTP_METHODS}")

    def _fn():
        with requests.Session() as sess:
            out = sess.request(
                method,
                url,
                params=params,
                data=data,
                headers=headers,
                cookies=cookies,
                auth=auth,  # type: ignore
                timeout=None if not timeout else timeout,
                allow_redirects=True,
                json=json,
            )
        return out.text, out.status_code

    text, status_code = exponential_backoff(foo=_fn, max_retries=max_retries, retry_delay=retry_delay)
    return (text, status_code), None  # type: ignore


programatic_actions_registry.register(
    call_api_requests,
    "call_api_requests",
    "Call an API using the requests library",
    outputs={
        "text": (0,),
        "status_code": (1,),
    },
)


# a few functions that do regex things


def regex_search(pattern: str, text: str) -> Tuple[List[str], Optional[Exception]]:
    """
    Perform a regex search on the text and get items in an array

    Args:
        pattern (str): The regex pattern to search for
        text (str): The text to search in

    Returns:
        Tuple[List[str], Optional[Exception]]: The list of items found
    """
    try:
        out = re.findall(pattern, text)
        return out, None
    except Exception as e:
        return [], e


programatic_actions_registry.register(
    regex_search,
    "regex_search",
    "Perform a regex search on the text and get items in an array",
    returns=["items"],
)


def regex_substitute(pattern: str, repl: str, text: str) -> Tuple[str, Optional[Exception]]:
    """
    Perform a regex substitution on the text and get the result

    Args:
        pattern (str): The regex pattern to search for
        repl (str): The replacement string
        text (str): The text to search in

    Returns:
        Tuple[str, Optional[Exception]]: The substituted text and the exception if there was one
    """
    try:
        out = re.sub(pattern, repl, text)
        return out, None
    except Exception as e:
        return "", e


programatic_actions_registry.register(
    regex_substitute,
    "regex_substitute",
    "Perform a regex substitution on the text and get the result",
    returns=["text"],
)


def json_translator(
    data: Union[str, Dict[str, str]],
    resolver: Dict[str, str],
    default: str = "",
    return_only_value: bool = False,
) -> Tuple[str, Optional[Exception]]:
    """
    This simple function takes a json string or a python dictionary and translates it to the required output defined by
    the `resolver`. It is a dictionary that tells the location of the output that you want and the target locations.
    Here is a simple example on how you can use this:

    .. code-block:: python

        >>> x = {
        ...   "a": {
        ...     "b": [1, 2, 3],
        ...     "c": {
        ...       "d": "hello",
        ...       "e": "world",
        ...     }
        ...   },
        ...   "f": "foo",
        ... }
        >>> resolver = {
        ...   "x": ["a", "b", 0],
        ...   "y": ["a", "c", "d"],
        ...   "z": ["f"],
        ... }
        >>> json_translator(x, resolver)
        >>> {
        ...   "x": 1,
        ...   "y": "hello",
        ...   "z": "foo",
        ... }

    Note:
        If you pass `return_only_value=True`, then the output will be the value of the first key in the resolver. See
        below for an example:

    .. code-block:: python

        >>> json_translator(x, resolver, return_only_value=True)
        >>> 1

    Args:
        data (Union[str, Dict[str, Any]]): The data to be processed
        resolver (Dict[str, str], optional): The resolver dictionary. Defaults to {}.
        default (str, optional): The default value to be returned if the resolver fails. Defaults to "".
        return_only_value (bool, optional): If True, only the value is returned. Defaults to False.

    Returns:
        Tuple[str, Optional[Exception]]: The output and the exception if any
    """
    if type(data) == str:
        data = json.loads(data)
    out = {}
    for k, v in resolver.items():
        if isinstance(v, dict):
            _temp = {}
            for k1, v1 in v.items():
                _temp[k1] = get_value_by_keys(data, v1)
            out[k] = _temp
        else:
            out[k] = get_value_by_keys(data, v)
    if return_only_value:
        out = next(iter(out.items()))[1]
    out = default or out
    return json.dumps(out), None


programatic_actions_registry.register(
    json_translator,
    "json_translator",
    "Extract a value from a JSON object using a list of keys",
    returns=["value"],
)
